import datetime
import email
import email.parser
import glob
import logging
import mailbox
import os
import re
import subprocess
import time
from urllib.parse import urlparse
from urllib.parse import urljoin
import warnings
from email.header import Header
from email.message import Message
from email.mime.text import MIMEText
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import requests
import yaml
from bs4 import BeautifulSoup

from config.config import CONFIG

project_directory = str(Path(os.path.abspath(__file__)).parent.parent)

logging.basicConfig(
    filename=project_directory + "/listserv.log",
    level=logging.INFO,
    format="%(asctime)s %(message)s",
)
logger = logging.getLogger(__name__)


class ListservMessageWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


class ListservListWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


class ListservArchiveWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


class ListservMessage:
    """
    Parameters
    ----------
    body
    subject
    fromname
    fromaddr
    toname
    toaddr
    date
    contenttype
    messageid

    Methods
    -------
    from_url
    get_header_from_html
    get_body_from_html
    get_header_from_listserv_file
    get_body_from_listserv_file
    get_name
    get_addr
    get_date
    remove_unwanted_header_content
    to_dict
    to_mbox

    Example
    -------
    msg = ListservMessage.from_url(
        list_name="3GPP_TSG_CT_WG6",
        url=url_message,
        fields="total",
    )
    """

    empty_header = {
        "subject": None,
        "fromname": None,
        "fromaddr": None,
        "toname": None,
        "toaddr": None,
        "date": "Mon, 11 Jan 1111 11:11:11",
        "contenttype": None,
    }

    def __init__(
        self,
        body: str,
        subject: str,
        fromname: str,
        fromaddr: str,
        toname: str,
        toaddr: str,
        date: str,
        contenttype: str,
        messageid: Optional[str] = None,
    ):
        self.body = body
        self.subject = subject
        self.fromname = fromname
        self.fromaddr = fromaddr
        self.toname = toname
        self.toaddr = toaddr
        self.date = date
        self.contenttype = contenttype

    @classmethod
    def from_url(
        cls,
        list_name: str,
        url: str,
        fields: str = "total",
        url_login: str = "https://list.etsi.org/scripts/wa.exe?LOGON",
        login: Optional[Dict[str, str]] = {"username": None, "password": None},
        session: Optional[str] = None,
    ) -> "ListservMessage":
        """
        Args:
        """
        if session is None:
            session = get_auth_session(url_login, **login)
        soup = get_website_content(url, session=session)
        if fields in ["header", "total"]:
            header = ListservMessage.get_header_from_html(soup)
        else:
            header = cls.empty_header
        if fields in ["body", "total"]:
            body = ListservMessage.get_body_from_html(list_name, url, soup)
        else:
            body = None
        return cls(body, **header)

    @classmethod
    def from_listserv_file(
        cls,
        list_name: str,
        file_path: str,
        header_start_line_nr: int,
        fields: str = "total",
    ) -> "ListservMessage":
        file = open(file_path, "r", errors="replace")
        fcontent = file.readlines()
        file.close()
        header_end_line_nr = cls.get_header_end_line_nr(
            fcontent, header_start_line_nr
        )
        if fields in ["header", "total"]:
            header = cls.get_header_from_listserv_file(
                fcontent, header_start_line_nr, header_end_line_nr
            )
        else:
            header = cls.empty_header
        if fields in ["body", "total"]:
            body = cls.get_body_from_listserv_file(
                fcontent, header_end_line_nr
            )
        else:
            body = None
        return cls(body, **header)

    @classmethod
    def get_header_end_line_nr(
        cls,
        content: List[str],
        header_start_line_nr: int,
    ) -> List[int]:
        """
        The header ends with the first empty line encountered.

        Args:
            content: The content of one LISTSERV-file.
        """
        for lnr, lcont in enumerate(content[header_start_line_nr:]):
            if len(lcont) <= 1:
                header_end_line_nr = header_start_line_nr + lnr
                break
        return header_end_line_nr

    @classmethod
    def get_header_from_listserv_file(
        cls,
        content: List[str],
        header_start_line_nr: int,
        header_end_line_nr: int,
    ) -> Dict[str, str]:
        """
        Args:
            content:
        """
        content = content[header_start_line_nr:header_end_line_nr]
        # collect important info from LISTSERV header
        header = {}
        for lnr in range(len(content)):
            line = content[lnr]
            # get header keyword and value
            if re.match(r"\S+:\s+\S+", line):
                key = line.split(":")[0]
                value = line.replace(key + ":", "").strip().rstrip("\n")
                # if not at the end of header
                if lnr < len(content) - 1:
                    # if header-keyword value is split over two lines
                    if not re.match(r"\S+:\s+\S+", content[lnr + 1]):
                        value += " " + content[lnr + 1].strip().rstrip("\n")
                header[key.lower()] = value
        
        header = cls.format_header_content(header)
        header = cls.remove_unwanted_header_content(header)
        return header

    @classmethod
    def get_body_from_listserv_file(
        cls,
        content: List[str],
        header_end_line_nr: int,
    ) -> str:
        """"""
        found = False
        # find body 'position' in file
        for line_nr, line in enumerate(content[header_end_line_nr:]):
            if "=" * 73 in line:
                body_end_line_nr = line_nr + header_end_line_nr
                found = True
                break
        if not found:
            body_end_line_nr = -1
        # get body content
        body = content[header_end_line_nr:body_end_line_nr]
        # remove empty lines and join into one string
        body = ("").join([line for line in body if len(line) > 1])
        return body

    @classmethod
    def get_header_from_html(cls, soup: BeautifulSoup) -> Dict[str, str]:
        """"""
        text = soup.find(
            "b",
            text=re.compile(r"^\bSubject\b"),
        ).parent.parent.parent.parent.text
        # collect important info from LISTSERV header
        header = {}
        for field in text.split("Parts/Attachments:")[0].splitlines():
            if len(field) == 0:
                continue
            field_name = field.split(":")[0].strip()
            field_body = field.replace(field_name + ":", "").strip()
            header[field_name.lower()] = field_body

        header = cls.format_header_content(header)
        header = cls.remove_unwanted_header_content(header)
        return header

    @staticmethod
    def get_body_from_html(
        list_name: str, url: str, soup: BeautifulSoup
    ) -> str:
        """"""
        try:
            url_root = ("/").join(url.split("/")[:-2])
            a_tags = soup.select(f'a[href*="A3="][href*="{list_name}"]')
            href_plain_text = [
                tag.get("href")
                for tag in a_tags
                if "Fplain" in tag.get("href")
            ][0]
            body_soup = get_website_content(
                urljoin(url_root, href_plain_text)
            )
            return body_soup.find("pre").text
        except Exception:
            logger.info(
                f"The message body of {url} which is part of the "
                f"list {list_name} could not be loaded."
            )
            return None

    @classmethod
    def format_header_content(cls, header: Dict[str, str]) -> Dict[str, str]:
        "Formats LISTSERV 16.5 header fields to mbox convention."
        # define formatting configuration
        formatting_header_conf = {
            "from": {
                "fromname": cls.get_name,
                "fromaddr": cls.get_addr,
            },
            "reply-to": {
                "toname": cls.get_name,
                "toaddr": cls.get_addr,
            },
            "date": {"date": cls.get_date},
            "content-type": {"contenttype": None},
        }
        # run through format settings
        for key_old, value in formatting_header_conf.items():
            # if header contains field
            if key_old in header.keys():
                for key_new, fct in formatting_header_conf[key_old].items():
                    if fct is None or header[key_old] is None:
                        header[key_new] = header[key_old]
                    else:
                        header[key_new] = fct(header[key_old])
            # if header misses field
            else:
                for key_new in formatting_header_conf[key_old].keys():
                    header[key_new] = cls.empty_header[key_new]
        return header

    @classmethod
    def remove_unwanted_header_content(
        cls, header: Dict[str, str]
    ) -> Dict[str, str]:
        for key in list(header.keys()):
            if key not in list(cls.empty_header.keys()):
                del header[key]
        return header

    @staticmethod
    def get_name(line: str) -> str:
        # get string in between < and >
        email_of_sender = re.findall(r"\<(.*)\>", line)
        if email_of_sender:
            # remove email_of_sender from line
            name = line.replace("<" + email_of_sender[0] + ">", "")
            # remove special characters
            name = re.sub(r"[^a-zA-Z0-9]+", " ", name)
        else:
            name = line
        return name.strip()

    @staticmethod
    def get_addr(line: str) -> Union[str, None]:
        # get string in between < and >
        email_addr = re.findall(r"\<(.*)\>", line)
        if email_addr:
            email_addr = email_addr[0].strip()
        else:
            email_addr = ListservMessage.empty_header["fromaddr"]
        return email_addr

    @staticmethod
    def get_date(line: str) -> str:
        line = (" ").join(line.split(" ")[:-1]).lstrip()
        # convert format to local version of date and time
        date_time_obj = datetime.datetime.strptime(
            line, "%a, %d %b %Y %H:%M:%S"
        )
        return date_time_obj.strftime("%c")
    
    @staticmethod
    def create_message_id(
        date: Union[str, None],
        from_address: Union[str, None],
    ) -> str:
        #if date is None:
        #    date = 'Wed Jan 11 11:11:11 1111'
        message_id = (".").join([date, from_address])
        # remove special characters
        message_id = re.sub(r"[^a-zA-Z0-9]+", "", message_id)
        return message_id

    def to_dict(self) -> Dict[str, str]:
        dic = {
            "Body": self.body,
            "Subject": self.subject,
            "FromName": self.fromname,
            "FromAddr": self.fromaddr,
            "ToName": self.toname,
            "ToAddr": self.toaddr,
            "Date": self.date,
            "ContentType": self.contenttype,
        }
        return dic

    def to_mbox(self, filepath: str, mode: str = "w"):
        """
        Safe mail list to .mbox files.
        """
        message_id = ListservMessage.create_message_id(
            self.date,
            self.fromaddr,
        )
        f = open(filepath, mode, encoding="utf-8")
        f.write("\n")
        # check that header was selected
        if self.subject is not None:
            f.write(f"From b'{self.fromaddr}' {self.date}\n")
            f.write(f"Content-Type: {self.contenttype}\n")
            f.write(f"MIME-Version: 1.0\n")
            f.write(f"In-Reply-To: {self.toname} <b'{self.toaddr}'>\n")
            f.write(f"From: {self.fromname} <b'{self.fromaddr}'>\n")
            f.write(f"Subject: b'{self.subject}\n")
            f.write(f"Message-ID: <{message_id}>'\n")
            f.write(f"Date: {self.date}'\n")
            f.write("\n")
        # check that body was selected
        if self.body is not None:
            f.write(self.body)
            f.write("\n")
        f.close()


class ListservList:
    """
    This class handles a single mailing list of a public archive in the
    LISTSERV 16.5 format.

    Parameters
    ----------
    name
        The of whom the list (e.g. 3GPP_COMMON_IMS_XFER, IEEESCO-DIFUSION, ...)
    source
        Contains the information of the location of the mailing list.
        It can be either an URL where the list or a path to the file(s).
    msgs
        List of ListservMessage objects

    Methods
    -------
    from_url
    from_messages
    from_listserv_files
    from_listserv_directories
    get_messages_from_url
    get_period_urls
    get_line_numbers_of_header_starts
    get_index_of_elements_in_selection
    to_dict
    to_pandas_dataframe
    to_mbox

    Example
    -------
    mlist = ListservList.from_url(
        "3GPP_TSG_CT_WG6",
        url="https://list.etsi.org/scripts/wa.exe?A0=3GPP_TSG_CT_WG6",
        select={
            "years": (2020, 2021),
            "months": "January",
            "weeks": [1,5],
            "fields": "header",
        },
    )
    """

    def __init__(
        self,
        name: str,
        source: Union[List[str], str],
        msgs: List[ListservMessage],
    ):
        self.name = name
        self.source = source
        self.messages = msgs

    def __len__(self) -> int:
        return len(self.messages)

    def __iter__(self):
        return iter(self.messages)

    def __getitem__(self, index) -> ListservMessage:
        return self.messages[index]

    @classmethod
    def from_url(
        cls,
        name: str,
        url: str,
        select: Optional[dict] = None,
        url_login: str = "https://list.etsi.org/scripts/wa.exe?LOGON",
        login: Optional[Dict[str, str]] = {"username": None, "password": None},
        session: Optional[str] = None,
    ) -> "ListservList":
        """
        Args:
            name: Name of the list of messages, e.g. '3GPP_TSG_SA_WG2_UPCON'
            url: URL to the LISTSERV list.
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
        """
        if session is None:
            session = get_auth_session(url_login, **login)
        if select is None:
            select = {"fields": "total"}
        elif "fields" not in list(select.keys()):
            select["fields"] = "total"
        msgs = cls.get_messages_from_url(name, url, select, session)
        return cls.from_messages(name, url, msgs)

    @classmethod
    def from_messages(
        cls,
        name: str,
        url: str,
        messages: List[Union[str, ListservMessage]],
        fields: str = "total",
        url_login: str = "https://list.etsi.org/scripts/wa.exe?LOGON",
        login: Optional[Dict[str, str]] = {"username": None, "password": None},
        session: Optional[str] = None,
    ) -> "ListservList":
        """
        Args:
            messages: Can either be a list of URLs to specific LISTSERV messages
                or a list of `ListservMessage` objects.
        """
        if not messages:
            # create empty ListservList for ListservArchive
            msgs = messages
        elif isinstance(messages[0], str):
            # create ListservList from message URLs
            if session is None:
                session = get_auth_session(url_login, **login)
            msgs = []
            for idx, url in enumerate(messages):
                msgs.append(
                    ListservMessage.from_url(
                        list_name=name,
                        url=url,
                        fields=fields,
                        session=session,
                    )
                )
        else:
            # create ListservList from list of ListservMessages
            msgs = messages
        return cls(name, url, msgs)

    @classmethod
    def from_listserv_directories(
        cls,
        name: str,
        directorypaths: List[str],
        filedsc: str = "*.LOG?????",
        select: Optional[dict] = None,
    ) -> "ListservList":
        """
        Args:
            name: Name of the list of messages, e.g. '3GPP_TSG_SA_WG2_UPCON'.
            directorypaths: List of directory paths where LISTSERV formatted
                messages are.
            filedsc: A description of the relevant files, e.g. *.LOG?????
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
        """
        _filepaths = []
        # run through directories and collect all filepaths
        for directorypath in directorypaths:
            _filepaths.append(
                get_paths_to_files_in_directory(directorypath, filedsc)
            )
        # flatten list of lists
        filepaths = [fp for li in _filepaths for fp in li]
        return cls.from_listserv_files(name, filepaths, select)

    @classmethod
    def from_listserv_files(
        cls,
        name: str,
        filepaths: List[str],
        select: Optional[dict] = None,
    ) -> "ListservList":
        """
        Args:
            name: Name of the list of messages, e.g. '3GPP_TSG_SA_WG2_UPCON'
            filepaths: List of file paths where LISTSERV formatted messages are.
                Such files can have a file extension of the form: *.LOG1405D
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
        """
        if select is None:
            select = {"fields": "total"}
        msgs = []
        for filepath in filepaths:
            # TODO: implement selection filter
            file = open(filepath, "r", errors="replace")
            fcontent = file.readlines()
            # get positions of all Emails in file
            header_start_line_nrs = cls.get_line_numbers_of_header_starts(
                fcontent
            )
            file.close()
            # run through all messages in file
            for msg_nr in header_start_line_nrs:
                msgs.append(
                    ListservMessage.from_listserv_file(
                        name,
                        filepath,
                        msg_nr,
                        select["fields"],
                    )
                )
        return cls(name, filepaths, msgs)

    @classmethod
    def get_messages_from_url(
        cls,
        name: str,
        url: str,
        select: Optional[dict] = None,
        session: Optional[dict] = None,
    ) -> List[ListservMessage]:
        """
        Generator that yields all messages within a certain period
        (e.g. January 2021, Week 5).

        Args:
            name: Name of the list of messages, e.g. '3GPP_TSG_SA_WG2_UPCON'
            url: URL to the LISTSERV list.
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
            session: AuthSession
        """
        if select is None:
            select = {"fields": "total"}
        msgs = []
        # run through periods
        for period_url in ListservList.get_period_urls(url, select):
            # run through messages within period
            for msg_url in ListservList.get_messages_urls(name, period_url):
                msgs.append(
                    ListservMessage.from_url(
                        name,
                        msg_url,
                        select["fields"],
                        session=session,
                    )
                )
                logger.info(f"Recorded the message {msg_url}.")
                # wait between loading messages, for politeness
                time.sleep(1)
        return msgs

    @classmethod
    def get_period_urls(
        cls, url: str, select: Optional[dict] = None
    ) -> List[str]:
        """
        All messages within a certain period
        (e.g. January 2021, Week 5).
        """
        # create dictionary with key indicating period and values the url
        periods, urls_of_periods = cls.get_all_periods_and_their_urls(url)

        if any(
            period in list(select.keys())
            for period in ["years", "months", "weeks"]
        ):
            for key, value in select.items():
                if key == "years":
                    cond = lambda x: int(re.findall(r"\d{4}", x)[0])
                elif key == "months":
                    cond = lambda x: x.split(" ")[0]
                elif key == "weeks":
                    cond = lambda x: int(x.split(" ")[-1])
                else:
                    continue

                periodquants = [cond(period) for period in periods]

                indices = ListservList.get_index_of_elements_in_selection(
                    periodquants,
                    urls_of_periods,
                    value,
                )

                periods = [periods[idx] for idx in indices]
                urls_of_periods = [urls_of_periods[idx] for idx in indices]
        return urls_of_periods

    @staticmethod
    def get_all_periods_and_their_urls(url: str) -> Tuple[List[str], List[str]]:
        url_root = ("/").join(url.split("/")[:-2])
        soup = get_website_content(url)
        periods = [list_tag.find("a").text for list_tag in soup.find_all("li")]
        urls_of_periods = [
            urljoin(url_root, list_tag.find("a").get("href"))
            for list_tag in soup.find_all("li")
        ]
        return periods, urls_of_periods

    @staticmethod
    def get_index_of_elements_in_selection(
        times: List[Union[int, str]],
        urls: List[str],
        filtr: Union[tuple, list, int, str],
    ) -> List[int]:
        """
        Filter out messages that where in a specific period. Period here is a set
        containing units of year, month, and week-of-month which can have the following
        example elements:
            - years: (1992, 2010), [2000, 2008], 2021
            - months: ["January", "July"], "November"
            - weeks: (1, 4), [1, 5], 2

        Args:
            times: A list containing information of the period for each
                group of ListservMessage.
            urls: Corresponding URLs of each group of ListservMessage of which the
                period info is contained in `times`.
            filtr: Containing info on what should be filtered.

        Returns:
            Indices of to the elements in `times`/`ursl`.
        """
        if isinstance(filtr, tuple):
            # filter year or week in range
            cond = lambda x: (np.min(filtr) <= x <= np.max(filtr))
        if isinstance(filtr, list):
            # filter in year, week, or month in list
            cond = lambda x: x in filtr
        if isinstance(filtr, int):
            # filter specific year or week
            cond = lambda x: x == filtr
        if isinstance(filtr, str):
            # filter specific month
            cond = lambda x: x == filtr
        return [idx for idx, time in enumerate(times) if cond(time)]

    @classmethod
    def get_messages_urls(cls, name: str, url: str) -> List[str]:
        """
        Args:
            name: Name of the `ListservList`
            url: URL to group of messages that are within the same period.

        Returns:
            List to URLs from which`ListservMessage` can be initialized.
        """
        url_root = ("/").join(url.split("/")[:-2])
        soup = get_website_content(url)
        a_tags = soup.select(f'a[href*="A2="][href*="{name}"]')
        if a_tags:
            a_tags = [
                urljoin(url_root, url.get("href"))
                for url in a_tags
            ]
        return a_tags

    @classmethod
    def get_line_numbers_of_header_starts(
        cls, content: List[str]
    ) -> List[int]:
        """
        By definition LISTSERV logs seperate new messages by a row
        of 73 equal signs.

        Args:
            content: The content of one LISTSERV-file.

        Returns:
            List of line numbers where header starts
        """
        return [
            line_nr for line_nr, line in enumerate(content) if "=" * 73 in line
        ]

    def to_conversationkg_dict(self) -> Dict[str, List[str]]:
        """
        Place all message in all lists into a dictionary of the form:
            dic = {
                "message_ID1": {
                    "body": ...,
                    "subject": ...,
                    ... ,
                }
                "message_ID2": {
                    "body": ...,
                    "subject": ...,
                    ... ,
                }
            }
        """
        # initialize dictionary
        dic = {}
        msg_nr = 0
        # run through messages
        for msg in self.messages:
            dic[f"ID{msg_nr}"] = msg.to_dict()
            msg_nr += 1
        return dic

    def to_dict(self) -> Dict[str, List[str]]:
        """
        Place all message into a dictionary of the form:
            dic = {
                "Subject": [messages[0], ... , messages[n]],
                .
                .
                .
                "ContentType": [messages[0], ... , messages[n]]
            }
        """
        # initialize dictionary
        dic = {}
        for key in list(self.messages[0].to_dict().keys()):
            dic[key] = []
        # run through messages
        for msg in self.messages:
            # run through message attributes
            for key, value in msg.to_dict().items():
                dic[key].append(value)
        return dic

    def to_pandas_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.to_dict())

    def to_mbox(self, dir_out: str, filename: Optional[str] = None):
        """
        Safe mail list to .mbox files.

        Args:
        """
        if filename is None:
            filepath = f"{dir_out}/{self.name}.mbox"
        else:
            filepath = f"{dir_out}/{filename}.mbox"
        logger.info(f"The list {self.name} is save at {filepath}.")
        first = True
        for msg in self.messages:
            if first:
                msg.to_mbox(filepath, mode="w")
                first = False
            else:
                msg.to_mbox(filepath, mode="a")


class ListservArchive(object):
    """
    This class handles a public mailing list archive that uses the
    LISTSERV 16.5 format.
    An archive is a list of ListservList elements.

    Parameters
    ----------
    name
        The of whom the archive is (e.g. 3GPP, IEEE, ...)
    url
        The URL where the archive lives
    lists
        A list containing the mailing lists as `ListservList` types

    Methods
    -------
    from_url
    from_mailing_lists
    from_listserv_directory
    get_lists
    get_sections
    to_dict
    to_pandas_dataframe
    to_mbox

    Example
    -------
    arch = ListservArchive.from_url(
        "3GPP",
        "https://list.etsi.org/scripts/wa.exe?",
        "https://list.etsi.org/scripts/wa.exe?HOME",
        select={
            "years": (2020, 2021),
            "months": "January",
            "weeks": [1,5],
            "fields": "header",
        },
    )
    """

    def __init__(self, name: str, url: str, lists: List[ListservList]):
        self.name = name
        self.url = url
        self.lists = lists

    def __len__(self):
        return len(self.lists)

    def __iter__(self):
        return iter(self.lists)

    def __getitem__(self, index):
        return self.lists[index]

    @classmethod
    def from_url(
        cls,
        name: str,
        url_root: str,
        url_home: str,
        select: Optional[dict] = None,
        url_login: str = "https://list.etsi.org/scripts/wa.exe?LOGON",
        login: Optional[Dict[str, str]] = {"username": None, "password": None},
        session: Optional[str] = None,
        instant_save: bool = True,
        only_mlist_urls: bool = True,
    ) -> "ListservArchive":
        """
        Create ListservArchive from a given URL.

        Args:
            name:
            url_root:
            url_home:
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
            url_login: URL to the login page
            login: login keys {"username": str, "password": str}
            session: if auth-session was already created externally
            instant_save: Boolean giving the choice to save a `ListservList` as
                soon as it is completely scraped or collect entire archive. The
                prior is recommended if a large number of mailing lists are
                scraped which can require a lot of memory and time.
            only_list_urls: Boolean giving the choice to collect only `ListservList`
                URLs or also their contents.
        """
        session = get_auth_session(url_login, **login)
        lists = cls.get_lists_from_url(
            url_root,
            url_home,
            select,
            session,
            instant_save,
            only_mlist_urls,
        )
        return cls.from_mailing_lists(
            name,
            url_root,
            lists,
            select,
            session,
            only_mlist_urls,
        )

    @classmethod
    def from_mailing_lists(
        cls,
        name: str,
        url_root: str,
        url_mailing_lists: Union[List[str], List[ListservList]],
        select: Optional[dict] = None,
        url_login: str = "https://list.etsi.org/scripts/wa.exe?LOGON",
        login: Optional[Dict[str, str]] = {"username": None, "password": None},
        session: Optional[str] = None,
        only_mlist_urls: bool = True,
    ) -> "ListservArchive":
        """
        Create ListservArchive from a given list of 'ListservList'.

        Args:
            name:
            url_root:
            url_mailing_lists:

        """
        if isinstance(url_mailing_lists[0], str) and only_mlist_urls is False:
            if session is None:
                session = get_auth_session(url_login, **login)
            lists = []
            for idx, url in enumerate(url_mailing_lists):
                lists.append(
                    ListservList.from_url(
                        name=idx,
                        url=url,
                        select=select,
                        session=session,
                    )
                )
        else:
            lists = url_mailing_lists
        return cls(name, url_root, lists)

    @classmethod
    def from_listserv_directory(
        cls,
        name: str,
        directorypath: str,
        folderdsc: str = "*",
        filedsc: str = "*.LOG?????",
        select: Optional[dict] = None,
    ) -> "ListservList":
        """
        Args:
            name: Name of the archive, e.g. '3GPP'.
            directorypath: Where the ListservArchive can be initialised.
            folderdsc: A description of the relevant folders
            filedsc: A description of the relevant files, e.g. *.LOG?????
            select: Selection criteria that can filter messages by:
                - content, i.e. header and/or body
                - period, i.e. written in a certain year, month, week-of-month
        """
        lists = []
        _dirpaths_to_lists = get_paths_to_dirs_in_directory(
            directorypath, folderdsc
        )
        # run through directories and collect all filepaths
        for dirpath in _dirpaths_to_lists:
            _filepaths = get_paths_to_files_in_directory(dirpath, filedsc)
            mlist = ListservList.from_listserv_files(
                dirpath.split("/")[-2],
                _filepaths,
                select,
            )
            lists.append(mlist)
        return cls(name, directorypath, lists)

    @staticmethod
    def get_lists_from_url(
        url_root: str,
        url_home: str,
        select: dict,
        session: Optional[str] = None,
        instant_save: bool = True,
        only_mlist_urls: bool = True,
    ) -> List[Union[ListservList, str]]:
        """
        Created dictionary of all lists in the archive.

        Args:

        Returns:
            archive_dict: the keys are the names of the lists and the value their url
        """
        archive = []
        # run through archive sections
        for url in list(
            ListservArchive.get_sections(url_root, url_home).keys()
        ):
            soup = get_website_content(url)
            a_tags_in_section = soup.select(
                f'a[href^="{urlparse(url).path}?A0="]',
            )

            mlist_urls = [
                urljoin(url_root, a_tag.get("href"))
                for a_tag in a_tags_in_section
            ]
            mlist_urls = list(set(mlist_urls))  # remove duplicates

            if only_mlist_urls:
                # collect mailing-list urls
                [
                    archive.append(mlist_url)
                    for mlist_url in mlist_urls
                    if len(ListservList.get_all_periods_and_their_urls(mlist_url)[1]) > 0
                ]

            else:
                # collect mailing-list contents
                for mlist_url in mlist_urls:
                    key = mlist_url.split("A0=")[-1]
                    mlist = ListservList.from_url(
                        name=key,
                        url=mlist_url,
                        select=select,
                        session=session,
                    )
                    if len(mlist) != 0:
                        if instant_save:
                            mlist.to_mbox(dir_out=CONFIG.mail_path)
                            archive.append(mlist.name)
                        else:
                            logger.info(f"Recorded the list {mlist.name}.")
                            archive.append(mlist)
        return archive

    def get_sections(url_root: str, url_home: str) -> int:
        """
        Get different sections of archive. On the website they look like:
        [3GPP] [3GPP–AT1] [AT2–CONS] [CONS–EHEA] [EHEA–ERM_] ...

        Returns:
            If sections exist, it returns their urls and names. Otherwise it returns
            the url_home.
        """
        soup = get_website_content(url_home)
        sections = soup.select(
            'a[href*="INDEX="][href*="p="]',
        )
        archive_sections_dict = {}
        if sections:
            for sec in sections:
                key = urljoin(url_root, sec.get("href"))
                value = sec.text
                if value in ["Next", "Previous"]:
                    continue
                archive_sections_dict[key] = value
            archive_sections_dict[re.sub(r'p=[0-9]+', 'p=1', key)] = 'FIRST'
        else:
            archive_sections_dict[url_home] = "Home"
        return archive_sections_dict

    def to_conversationkg_dict(self) -> Dict[str, List[str]]:
        """
        Place all message in all lists into a dictionary of the form:
            dic = {
                "message_ID1": {
                    "body": ...,
                    "subject": ...,
                    ... ,
                }
                "message_ID2": {
                    "body": ...,
                    "subject": ...,
                    ... ,
                }
            }
        """
        # initialize dictionary
        dic = {}
        msg_nr = 0
        # run through lists
        for mlist in self.lists:
            # run through messages
            for msg in mlist.messages:
                dic[f"ID{msg_nr}"] = msg.to_dict()
                msg_nr += 1
        return dic

    def to_dict(self) -> Dict[str, List[str]]:
        """
        Place all message in all lists into a dictionary of the form:
            dic = {
                "Subject": [messages[0], ... , messages[n]],
                .
                .
                .
                "ListName": [messages[0], ... , messages[n]]
            }
        """
        # initialize dictionary
        dic = {}
        for key in list(self.lists[0].messages[0].to_dict().keys()):
            dic[key] = []
        dic["ListName"] = []
        # run through lists
        for mlist in self.lists:
            # run through messages
            for msg in mlist.messages:
                # run through message attributes
                for key, value in msg.to_dict().items():
                    dic[key].append(value)
                dic["ListName"].append(mlist.name)
        return dic

    def to_pandas_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.to_dict())

    def to_mbox(self, dir_out: str):
        """
        Save Archive content to .mbox files
        """
        for llist in self.lists:
            llist.to_mbox(dir_out)


def get_auth_session(
    url_login: str, username: str, password: str
) -> requests.Session:
    """ Create AuthSession """
    # ask user for login keys
    username, password = get_login_from_terminal(username, password)
    if username is None or password is None:
        # continue without authentication
        return None
    else:
        # Start the AuthSession
        session = requests.Session()
        # Create the payload
        payload = {
            "LOGIN1": "",
            "Y": username,
            "p": password,
            "X": "",
        }
        # Post the payload to the site to log in
        session.post(url_login, data=payload)
        return session


def get_login_from_terminal(
    username: Union[str, None],
    password: Union[str, None],
    file_auth: str = project_directory + "/config/authentication.yaml",
) -> Tuple[Union[str, None]]:
    """
    Get login key from user during run time if 'username' and/or 'password' is 'None'.
    Return 'None' if no reply within 15 sec.
    """
    if username is None or password is None:
        record = True
    else:
        record = False
    if username is None:
        username = ask_for_input("Enter your Email: ")
    if password is None:
        password = ask_for_input("Enter your Password: ")
    if record and isinstance(username, str) and isinstance(password, str):
        loginkey_to_file(username, password, file_auth)
    return username, password


def ask_for_input(request: str) -> Union[str, None]:
    timeout = 15
    end_time = time.time() + timeout
    while time.time() < end_time:
        reply = input(request)
        try:
            assert isinstance(reply, str)
            break
        except Exception:
            reply = None
            continue
    return reply


def loginkey_to_file(
    username: str,
    password: str,
    file_auth: str,
) -> None:
    """ Safe login key to yaml """
    file = open(file_auth, "w")
    file.write(f"username: '{username}'\n")
    file.write(f"password: '{password}'")
    file.close()


def get_website_content(
    url: str,
    session: Optional[requests.Session] = None,
) -> BeautifulSoup:
    """ Get HTML code from website """
    # TODO: include option to change BeautifulSoup args
    if session is None:
        sauce = requests.get(url)
        assert sauce.status_code == 200
        soup = BeautifulSoup(sauce.content, "html.parser")
    else:
        sauce = session.get(url)
        soup = BeautifulSoup(sauce.text, "html.parser")
    return soup


def get_paths_to_files_in_directory(
    directory: str, file_dsc: str = "*"
) -> List[str]:
    """ Get paths of all files matching file_dsc in directory """
    template = f"{directory}{file_dsc}"
    file_paths = glob.glob(template)
    return file_paths


def get_paths_to_dirs_in_directory(
    directory: str, folder_dsc: str = "*"
) -> List[str]:
    """ Get paths of all directories matching file_dsc in directory """
    template = f"{directory}{folder_dsc}"
    dir_paths = glob.glob(template)
    # normalize directory paths
    dir_paths = [dir_path + "/" for dir_path in dir_paths]
    return dir_paths



import math
import re

import pandas as pd

email_regex = r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+"
domain_regex = r"[@]([a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)$"


def extract_email(from_field):
    """
    Returns an email address from a string.
    """
    match = re.search(email_regex, from_field)

    if match is not None:
        return match[0].lower()

    else:
        return None


def extract_domain(from_field):
    """
    Returns the domain of an email address from a string.
    """
    match = re.search(email_regex, from_field)

    if match is not None:
        return re.search(domain_regex, match[0])[1]

    else:
        return None



from dbigbang.utils import clean_message


class Thread:
    def __init__(self, root, known_root=True):
        """Form a thread object. root: the node of the message that start the thread
        known_root: indicator whether the root node is in our data set
        """
        self.root = root
        self.known_root = known_root

    def get_root(self):
        """Return the root node."""
        return self.root

    def get_num_messages(self):
        """Return the number of messages in the thread"""
        if self.known_root:
            return self.root.properties()[0]
        else:
            return 1 + self.root.get_successors()[0].properties()[0]

    def get_num_people(self):
        """Return the number of people in the thread"""
        if self.known_root:
            return len(self.root.properties()[1])
        else:
            return len(self.root.get_successors()[0].properties()[1])

    def get_duration(self):
        """Return the time duration of the thread"""
        if self.known_root:
            r = self.root
        else:
            r = self.root.get_successors()[0]
        _l = r.properties()[2]
        l_time = sorted([i.data["Date"] for i in _l])
        return l_time[len(_l) - 1] - r.data["Date"]

    def get_leaves(self):
        if self.known_root:
            r = self.root
        else:
            r = self.root.get_successors()[0]
        return r.properties()[2]

    def get_not_leaves(self):
        if self.known_root:
            r = self.root
        else:
            r = self.root.get_successors()[0]
        return r.properties()[4]

    def get_content(self):
        if self.known_root:
            r = self.root
        else:
            r = self.root.get_successors()[0]
        return r.properties()[3]


class Node:
    def __init__(self, ID, data=None, parent=None):
        """
        Form a Node object.
        ID: Message ID, data: Information about that message, parent: the
        message's reply-to
        """
        self.id = ID
        self.parent = parent
        self.successors = list()
        self.data = data
        self.processed = False
        self.prop = dict()

    def add_successor(self, successor: list):
        """Add a node which has a message that is a reply to this node"""
        self.successors.append(successor)

    def get_id(self):
        """Return message ID"""
        return self.id

    def get_successors(self):
        """Return a list of nodes of messages which are replies to this node"""
        return self.successors

    def get_data(self):
        """Return the Information about this message"""
        return self.data

    def get_parent(self):
        """Return Information in the data set about this message"""
        return self.parent

    def properties(self):
        """Return various properties about the tree with this node as root."""
        visited = set()
        seen_email = set()
        leaves = []
        not_leaves = []
        content = []

        def explore(node):
            num = 1
            seen_email.add(node.data["From"])
            visited.add(node)
            content.append(clean_message(node.data["Body"]))
            if len(node.get_successors()) == 0:
                leaves.append(node)
            else:
                not_leaves.append(node)
            for s in node.get_successors():
                if s not in visited:
                    seen_email.add(s.data["From"])
                    num += explore(s)
            return num

        if not self.processed:
            num_nodes = explore(self)
            self.prop["num_nodes"] = num_nodes
            self.prop["email_adrress"] = seen_email
            self.prop["leaves"] = leaves
            self.prop["content"] = content
            self.prop["not_leaves"] = not_leaves
            self.processed = True
        return [
            self.prop["num_nodes"],
            self.prop["email_adrress"],
            self.prop["leaves"],
            self.prop["content"],
            self.prop["not_leaves"],
        ]



import matplotlib.patches as patches
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


def stack(
    df,
    partition=None,
    smooth=1,
    figsize=(12.5, 7.5),
    time=True,
    cm=plt.cm.Set1,
):
    """
    Plots a stackplot based on a dataframe. Includes support for partitioning
    and convolution.

    df - a dataframe
    partition - a (dictionary or list) of lists of columns of df
              - if dictionary, keys are used as labels
    smooth - an integer amount of convolution
    """

    # partitioning logic
    if partition is None:
        # every column stacked separately
        partition = [[c] for c in df.columns]

    if isinstance(partition, list):
        partition = enumerate(partition)
    else:
        partition = list(partition.items())

    # if an item in the portition is not in the dataframe,
    # clean it up
    partition = [
        (kv[0], [i for i in kv[1] if i in df.columns]) for kv in partition
    ]
    # remove empty partitions
    partition = [kv for kv in partition if kv[1]]

    # convolution logic
    convolve_array = np.array([1.0 / smooth] * smooth)

    part_d = [
        np.convolve(df.as_matrix(part[1]).sum(1), convolve_array, mode="same")
        for part in partition
    ]

    d = np.row_stack(part_d)

    fig, ax = plt.subplots(figsize=figsize)

    # colors = [cm(float(k) / len(partition)) for k in range(len(partition))]
    colors = cm(np.linspace(0, 1, num=len(partition)))

    stacks = ax.stackplot(df.index, d, linewidth=0, colors=colors)

    # time axes
    if time:
        fig.axes[0].xaxis_date()

    # legend
    # make proxy artists
    proxy_rects = [
        patches.Rectangle((0, 0), 1, 1, fc=pc.get_facecolor()[0])
        for pc in stacks
    ]
    # make the legend
    ax.legend(proxy_rects, [kv[0] for kv in partition])

    plt.show()


# modified from http://sociograph.blogspot.com/2012/11/visualizing-adjacency-matrices-in-python.html
def draw_adjacency_matrix(
    G, node_order=None, partitions=[], colors=[], cmap="Greys", figsize=(6, 6)
):
    """
    - G is a networkx graph
    - node_order (optional) is a list of nodes, where each node in G
          appears exactly once
    - partitions is a list of node lists, where each node in G appears
          in exactly one node list
    - colors is a list of strings indicating what color each
          partition should be
    If partitions is specified, the same number of colors needs to be
    specified.
    """
    adjacency_matrix = nx.to_numpy_matrix(
        G, dtype=np.bool, nodelist=node_order
    )

    plt.figure(figsize=figsize)  # in inches
    plt.imshow(adjacency_matrix, cmap=cmap, interpolation="none")

    # The rest is just if you have sorted nodes by a partition and want to
    # highlight the module boundaries
    assert len(partitions) == len(colors)
    ax = plt.gca()
    current_idx = 0
    for partition, color in zip(partitions, colors):
        # for module in partition:
        ax.add_patch(
            patches.Rectangle(
                (current_idx, current_idx),
                len(partition),  # Width
                len(partition),  # Height
                facecolor="none",
                edgecolor=color,
                linewidth="1",
            )
        )
        current_idx += len(partition)



import email
import email.header
import email.parser
import gzip
import logging
import mailbox
import os
import re
import time
import urllib.error
import urllib.parse
import urllib.request
from email.mime.text import MIMEText

import dateutil
from bs4 import BeautifulSoup

import dbigbang.mailman

from . import parse


class W3cMailingListArchivesParser(email.parser.Parser):
    """
    A subclass of email.parser.Parser that parses the HTML of single-message
    archive pages generated by W3C's mail archives system, based on Hypermail.
    """

    parse = None
    # doesn't implement the file version

    # TODO: ignore spam (has separate error message in w3c archives)
    def parsestr(self, text, headersonly=False):
        """
        Takes the full HTML of a single message page; returns an email Message
        as an mboxMessage, with appropriate From separator line.

        headersonly is not supported. Not all headers are being parsed yet.
        """
        soup = BeautifulSoup(text, "html.parser")
        body = self._text_for_selector(soup, "#body")
        msg = MIMEText(body, "plain", "utf-8")

        from_text = self._parse_dfn_header(
            self._text_for_selector(soup, "#from")
        )
        from_name = from_text.split("<")[0].strip()
        from_address = self._text_for_selector(soup, "#from a")

        from_addr = email.utils.formataddr(
            (from_name, email.header.Header(from_address).encode())
        )
        msg["From"] = from_addr

        subject = self._text_for_selector(soup, "h1")
        msg["Subject"] = subject

        message_id = self._parse_dfn_header(
            self._text_for_selector(soup, "#message-id")
        )
        msg["Message-ID"] = message_id.strip()

        message_date = self._parse_dfn_header(
            self._text_for_selector(soup, "#date")
        )
        msg["Date"] = message_date.strip()

        message_to = self._parse_dfn_header(
            self._text_for_selector(soup, "#to")
        )
        if message_to:
            msg["To"] = message_to.strip()

        message_cc = self._parse_dfn_header(
            self._text_for_selector(soup, "#cc")
        )
        if message_cc:
            msg["Cc"] = message_cc.strip()

        in_reply_to_pattern = re.compile('<!-- inreplyto="(.+?)"')
        match = in_reply_to_pattern.search(str(text))
        if match:
            msg["In-Reply-To"] = "<" + match.groups()[0] + ">"

        mbox_message = mailbox.mboxMessage(msg)
        mbox_message.set_from(
            email.header.Header(from_address).encode(),
            email.utils.parsedate(message_date),
        )

        return mbox_message

    def _parse_dfn_header(self, header_text):
        header_texts = str(header_text).split(":", 1)
        if len(header_texts) == 2:
            return header_texts[1]
        else:
            logging.debug("Split failed on %s", header_text)
            return ""

    def _text_for_selector(self, soup, selector):
        results = soup.select(selector)
        if results:
            result = results[0].get_text(strip=True)
        else:
            result = ""
            logging.debug("No matching text for selector %s", selector)

        return str(result)


def collect_from_url(url, base_arch_dir="archives", notes=None):
    """
    Collects W3C mailing list archives from a particular mailing list URL.

    Logs an error and returns False if no messages can be collected.
    """
    list_name = dbigbang.mailman.get_list_name(url)
    logging.info("Getting W3C list archive for %s", list_name)

    try:
        response = urllib.request.urlopen(url)
        response_url = response.geturl()
        html = response.read()
        soup = BeautifulSoup(html, "html.parser")
    except urllib.error.HTTPError:
        logging.exception("Error in loading W3C list archive page for %s", url)
        return False

    try:
        time_period_indices = list()
        rows = soup.select("tbody tr")
        for row in rows:
            link = row.select("td:nth-of-type(1) a")[0].get("href")
            logging.info("Found time period archive page: %s", link)
            time_period_indices.append(link)
    except Exception:
        logging.exception("Error in parsing list archives for %s", url)
        return False

    # directory for downloaded files
    arc_dir = dbigbang.mailman.archive_directory(base_arch_dir, list_name)
    dbigbang.mailman.populate_provenance(
        directory=arc_dir,
        list_name=list_name,
        list_url=url,
        notes=notes,
    )

    for link in time_period_indices:

        try:
            link_url = urllib.parse.urljoin(response_url, link)
            response = urllib.request.urlopen(link_url)
            html = response.read()
            soup = BeautifulSoup(html, "html.parser")
        except urllib.error.HTTPError:
            logging.exception("Error in loading: %s", link_url)
            return False

        end_date_string = (
            soup.select("#end")[0].parent.parent.select("em")[0].get_text()
        )
        end_date = dateutil.parser.parse(end_date_string)
        year_month_mbox = end_date.strftime("%Y-%m") + ".mbox"
        mbox_path = os.path.join(arc_dir, year_month_mbox)

        # looks like we've already downloaded this timeperiod
        if os.path.isfile(mbox_path):
            logging.info("Looks like %s already exists, moving on.", mbox_path)
            continue
        logging.info("Downloading messages to archive to %s.", mbox_path)

        message_links = list()
        messages = list()

        anchors = soup.select("div.messages-list a")
        for anchor in anchors:
            if anchor.get("href"):
                message_url = urllib.parse.urljoin(
                    link_url, anchor.get("href")
                )
                message_links.append(message_url)

        for message_link in message_links:
            response = urllib.request.urlopen(message_link)
            html = response.read()

            message = W3cMailingListArchivesParser().parsestr(html)
            message.add_header("Archived-At", "<" + message_link + ">")
            messages.append(message)
            time.sleep(1)  # wait between loading messages, for politeness

        mbox = mailbox.mbox(mbox_path)
        mbox.lock()

        try:
            for message in messages:
                mbox.add(message)
            mbox.flush()
        finally:
            mbox.unlock()

        logging.info("Saved %s", year_month_mbox)

    # assumes all archives were downloaded if no exceptions have been thrown
    provenance = dbigbang.mailman.access_provenance(arc_dir)
    provenance["complete"] = True
    dbigbang.mailman.update_provenance(arc_dir, provenance)



import re

import numpy as np
import pandas as pd

namesID = dict()
emailsID = dict()
allID = dict()
nameRegex = re.compile(r"(\(.*\))")
currID = 1


def getID(name, email):
    """Get ID from a name and email."""

    global currID
    global emailsID
    global namesID
    nameID = False
    emailID = False
    if name is not None:
        if name in namesID:
            nameID = namesID[name]
    if email is not None:
        if email in emailsID:
            emailID = emailsID[email]

    if (not emailID) and (not nameID):
        store(currID, name, email)
        currID += 1
        return currID

    if not emailID:
        store(nameID, name, email)
        return nameID

    if not nameID:
        store(emailID, name, email)
        return emailID

    if emailID != nameID:
        # print("ID MISMATCH! " + email + " " + name)
        store(nameID, name, email)
    else:
        if emailID not in allID:
            store(emailID, name, email)

    return nameID


def store(id, name, email):
    """Store name and email by ID."""

    if id not in allID:
        allID[id] = {"emails": list(), "names": list()}
    fullID = allID[id]
    namesID[name] = id
    emailsID[email] = id
    fullID["names"].append(name)
    fullID["emails"].append(email)


def name_for_id(id):
    """Return name by ID."""

    if id in allID:
        if "names" in allID[id] and len(allID[id]["names"]) > 0:
            return allID[id]["names"][0]
    return "UNKNOWN " + str(id)


def entity_resolve(row, emailCol, nameCol):
    """Return a row with name and email by ID."""

    emailAddress = row[emailCol].upper()
    emailAddress = emailAddress.replace(" AT ", "@")

    match = nameRegex.search(emailAddress)
    name = None

    if match is not None:
        name = match.group(0)  # unused info for now
        emailAddress = emailAddress.replace(name, "")
        name = name.replace("(", "")
        name = name.replace(")", "")
    if nameCol is not None:
        name = row[nameCol].upper()
    row["Person-ID"] = getID(name, emailAddress)
    return row



from datetime import datetime
from time import mktime

import git
import networkx as nx
import numpy as np
import pandas as pd
from git import *
from git import Repo

from config.config import CONFIG

from . import utils
from .entity_resolution import entity_resolve

ALL_ATTRIBUTES = (
    CONFIG.all_attributes
)  # ["HEXSHA", "Committer Name", "Committer Email", "Commit Message", "Time", "Parent Commit", "Touched File"]


def cache_fixer(r):
    """Adds info from row to graph."""
    r["Touched File"] = [x.strip() for x in r["Touched File"][1:-1].split(",")]
    r["Time"] = pd.to_datetime(r["Time"])
    return r


class GitRepo(object):
    """
    Store a git repository given the address to that repo relative to this file.

    It returns the data in many forms.
    """

    def __init__(self, name, url=None, attribs=ALL_ATTRIBUTES, cache=None):
        """
        Index a Pandas DataFrame object by time.

        That stores the raw form of the repo's commit data as a table.

        Each row in this table  is a commit.

        And each column represents an attribute of that commit:
        (eg.: time, message, commiter name, committer email, commit hexsha).
        """

        self._commit_data = None
        self.url = url
        self.repo = None
        self.name = name

        if cache is None:
            self.repo = Repo(url)
            self.populate_data(ALL_ATTRIBUTES)
        else:
            cache = cache.apply(cache_fixer, axis=1)
            cache.set_index(cache["Time"])
            self._commit_data = cache

            missing = list()
            cols = self.commit_data.columns
            for attr in attribs:
                if attr not in cols and str(attr) not in cols:
                    missing.append(attr)

            if len(missing) > 0:
                print(
                    (
                        "There were "
                        + str(len(missing))
                        + " missing attributes: "
                    )
                )
                print(missing)

        if "Committer Name" in attribs and "Committer Email" in attribs:
            self._commit_data["Person-ID"] = None
            self._commit_data = self._commit_data.apply(
                lambda row: entity_resolve(
                    row, "Committer Email", "Committer Name"
                ),
                axis=1,
            )

    def gen_data(self, repo, raw):
        """Generate data to repo."""

        if not repo.active_branch.is_valid():
            print(("Found an empty repo: " + str(self.name)))
            return
        first = repo.commit()
        commit = first
        firstHexSha = first.hexsha
        generator = git.Commit.iter_items(repo, firstHexSha)

        if "Touched File" in raw:
            print(
                "WARNING: Currently going through file diffs. This will take a very long time (1 minute per 3000 commits.) We suggest using a small repository."
            )
        for commit in generator:
            try:
                if "Touched File" in raw:
                    diff_list = list()
                    for diff in commit.diff(commit.parents[0]):
                        if diff.b_blob:
                            diff_list.append(diff.b_blob.path)
                        else:
                            diff_list.append(diff.a_blob.path)
                    raw["Touched File"].append(diff_list)

                if "Committer Name" in raw:
                    raw["Committer Name"].append(commit.committer.name)

                if "Committer Email" in raw:
                    raw["Committer Email"].append(commit.committer.email)

                if "Commit Message" in raw:
                    raw["Commit Message"].append(commit.message)

                if (
                    "Time" in raw or True
                ):  # TODO: For now, we always ask for the time
                    raw["Time"].append(
                        pd.to_datetime(commit.committed_date, unit="s")
                    )

                if "Parent Commit" in raw:
                    raw["Parent Commit"].append(
                        [par.hexsha for par in commit.parents]
                    )

                if "HEXSHA" in raw:
                    raw["HEXSHA"].append(commit.hexsha)
            except LookupError:
                print("failed to add a commit because of an encoding error")

    def populate_data(self, attribs=ALL_ATTRIBUTES):
        """Populate data."""
        raw = dict()
        for attrib in attribs:
            raw[attrib] = list()
        repo = self.repo
        self.gen_data(repo, raw)
        print((type(raw["Time"])))

        time_index = pd.DatetimeIndex(raw["Time"]).to_period("H")
        time_index = utils.add_freq(time_index, freq=None)
        self._commit_data = pd.DataFrame(raw, index = time_index);

    def by_committer(self):
        """Return commit data grouped by commiter."""
        return self.commit_data.groupby("Committer Name").size().order()

    def commits_per_day(self):
        """Return commits grouped by day."""
        ans = self.commit_data.groupby(self.commit_data.index).size()
        ans = ans.resample("D", how=np.sum)
        return ans

    def commits_per_week(self):
        """Return commits grouped by week."""
        ans = self.commits_per_day()
        ans = ans.resample("W", how=np.sum)
        return ans

    def commits_per_day_full(self):
        """Return commits grouped by day and by commiter."""
        ans = self.commit_data.groupby(
            [self.commit_data.index, "Committer Name"]
        ).size()
        return ans

    @property
    def commit_data(self):
        """Return commit data."""
        return self._commit_data

    def commits_for_committer(self, committer_name):
        """Return commits for committer given the commiter name."""
        full_info = self.commit_data

        df = full_info.loc[full_info["Committer Name"] == committer_name]
        df = df.groupby([df.index]).size()
        df = df.resample("D", how=np.sum, axis=0)

        return df

    def merge_with_repo(self, other):
        """Append commit to a repo."""
        # TODO: What if commits have the same time?
        self._commit_data = self.commit_data.append(other.commit_data)


class MultiGitRepo(GitRepo):
    """Repos must have a "Repo Name" column."""

    def __init__(self, repos, attribs=ALL_ATTRIBUTES):
        self._commit_data = repos[0].commit_data.copy(deep=True)
        for i in range(1, len(repos)):
            self.merge_with_repo(repos[i])



import datetime
import math
from collections import Counter
from pprint import pprint as pp

import networkx as nx
import numpy as np
import pandas
import pytz

import dbigbang.parse as parse
import dbigbang.process as process


def messages_to_reply_graph(messages):
    """Return a graph given messages."""

    G = nx.DiGraph()

    for m in messages:
        mid = parse.clean_mid(m.get("Message-ID"))

        G.add_node(mid)
        G.node[mid]["From"] = m.get("From")
        # G.node[mid]['Date'] = m.get('Date')
        # G.node[mid]['Message'] = m.get('Message')

        # references should be recoverable from in-reply-to structure
        # if 'References' in d:
        #    G.add_edge(mid,d['References'])

        if "In-Reply-To" in m:
            G.add_edge(mid, parse.clean_mid(m.get("In-Reply-To")))

    return G


def messages_to_interaction_graph(messages, verbose=False, clean=True):
    """Return a interactable graph given messages."""

    IG = nx.DiGraph()

    from_dict = {}
    sender_counts = {}
    reply_counts = {}

    if not isinstance(messages, pandas.core.frame.DataFrame):
        df = process.messages_to_dataframe(messages)
    else:
        df = messages

    for m in df.iterrows():
        m_from = m[1]["From"]
        if clean:
            m_from = parse.clean_from(m_from)

        from_dict[m[0]] = m_from
        sender_counts[m_from] = sender_counts.get(m_from, 0) + 1
        # the necessity of this initialization step may be dubious
        reply_counts[m_from] = reply_counts.get(m_from, {})
        IG.add_node(m_from)

    for sender, count in list(sender_counts.items()):
        IG.nodes[sender]["sent"] = count

    replies = [m for m in df.iterrows() if m[1]["In-Reply-To"] is not None]

    for m in replies:
        m_from = m[1]["From"]

        if clean:
            m_from = parse.clean_from(m_from)

        reply_to_mid = m[1]["In-Reply-To"]

        if reply_to_mid in from_dict:
            m_to = from_dict[reply_to_mid]
            reply_counts[m_from][m_to] = reply_counts[m_from].get(m_to, 0) + 1
        else:
            if verbose:
                print(reply_to_mid + " not in archive")

    for m_from, edges in list(reply_counts.items()):
        for m_to, count in list(edges.items()):
            IG.add_edge(m_from, m_to, weight=count)

    return IG


def interaction_graph_to_matrix(dg):
    """Turn an interaction graph into a weighted edge matrix."""
    nodes = dg.nodes()

    n_nodes = len(nodes)

    # n x n where n is number of nodes
    matrix = np.zeros([n_nodes, n_nodes])

    for m_from, m_to, data in dg.edges(data=True):
        i = nodes.index(m_from)
        j = nodes.index(m_to)

        matrix[i, j] = data["weight"]

    return matrix


def ascendancy(am):
    """
    Ulanowicz ecosystem health measures
    Input is weighted adjacency matrix.
    """
    # should these be normalized?!?!
    # output rates
    s0 = np.tile(np.sum(am, 0).T, (am.shape[0], 1))
    # input rates
    s1 = np.tile(np.sum(am, 1).T, (am.shape[1], 1)).T

    logs = np.nan_to_num(np.log(am * np.sum(am) / (s0 * s1)))

    # ascendancy!
    A = np.sum(am * logs)

    return A


def capacity(am):
    """Return the capacity given a adjacency matrix."""
    # total system throughput
    tst = np.sum(am)

    logs = np.nan_to_num(np.log(am / tst))

    return -np.sum(am * logs)


def overhead(am):
    """Return overhead given a adjacency matrix."""
    # could be more efficient...
    return capacity(am) - ascendancy(am)


""" copied from process.py may have bugs"""


def compute_ascendancy(messages, duration=50):
    """Compute ascendancy given messages."""

    print("compute ascendancy")
    dated_messages = {}

    for m in messages:
        d = parse.get_date(m)

        if d is not None and d < datetime.datetime.now(pytz.utc):
            o = d.toordinal()
            dated_messages[o] = dated_messages.get(o, [])
            dated_messages[o].append(m)

    days = [k for k in list(dated_messages.keys())]
    day_offset = min(days)
    epoch = max(days) - min(days)

    ascendancy = np.zeros([max(days) - min(days) + 1])
    capacity = np.zeros(([max(days) - min(days) + 1]))

    for i in range(epoch):
        min_d = min(days) + i
        max_d = min_d + duration

        block_messages = []

        for d in range(min_d, max_d):
            block_messages.extend(dated_messages.get(d, []))

        b_IG = messages_to_interaction_graph(block_messages)
        b_matrix = interaction_graph_to_matrix(b_IG)

        ascendancy[min_d - day_offset] = ascendancy(b_matrix)
        capacity[min_d - day_offset] = capacity(b_matrix)

    return ascendancy, capacity



import networkx as nx
import pandas as pd


def labeled_blockmodel(g, partition):
    """
    Perform blockmodel transformation on graph *g*
    and partition represented by dictionary *partition*.
    Values of *partition* are used to partition the graph.
    Keys of *partition* are used to label the nodes of the
    new graph.
    """
    new_g = nx.quotient_graph(g, list(partition.values()), relabel=True)
    labels = dict(enumerate(partition.keys()))
    new_g = nx.relabel_nodes(new_g, labels)

    return new_g


def repartition_dataframe(df, partition):
    """
    Create a new dataframe with the same index as
    argument dataframe *df*, where columns are
    the keys of dictionary *partition*.
    The data of the returned dataframe are the
    combinations of columns listed in the keys of
    *partition*
    """
    df2 = pd.DataFrame(index=df.index)

    for k, v in list(partition.items()):
        df2[k] = df[v[0]]

        for i in range(len(v) - 1):
            df2[k] = df2[k] + df[v[i]]

    return df2


def get_common_head(str1, str2, delimiter=None):
    try:
        if str1 is None or str2 is None:
            return "", 0
        else:
            # this is ugly control flow clean it
            if delimiter is not None:

                dstr1 = str1.split(delimiter)
                dstr2 = str2.split(delimiter)

                for i in range(len(dstr1)):
                    if dstr1[:i] != dstr2[:i]:
                        # print list1[:i], list2[:i]
                        return delimiter.join(dstr1[: i - 1]), i - 1
                return str1, i
            else:
                for i in range(len(str1)):
                    if str1[:i] != str2[:i]:
                        # print list1[:i], list2[:i]
                        return str1[: i - 1], i - 1
                return str1[:i], i

        return "", 0
    except Exception as e:
        print(e)
        return "", 0


def get_common_foot(str1, str2, delimiter=None):
    head, ln = get_common_head(str1[::-1], str2[::-1], delimiter=delimiter)

    return head[::-1], ln


# turn these into automated tests
# print get_common_head('abcdefghijklmnop','abcde12345')
# print get_common_head('abcdefghijklmnop',None)
# print get_common_head('abcde\nfghijk\nlmnop\nqrst','abcde\nfghijk\nlmnopqr\nst',delimiter='\n')


def remove_quoted(mess):
    message = [
        line
        for line in mess.split("\n")
        if len(line) != 0 and line[0] != ">" and line[-6:] != "wrote:"
    ]
    new = "\n".join(message)
    return new


## remove this when clean_message is added to generic libraries
def clean_message(mess):
    if mess is None:
        mess = ""

    mess = remove_quoted(mess)

    mess = mess.strip()

    return mess


# From here:
# https://stackoverflow.com/questions/46217529/pandas-datetimeindex-frequency-is-none-and-cant-be-set
def add_freq(idx, freq=None):
    """Add a frequency attribute to idx, through inference or directly.

    Returns a copy.  If `freq` is None, it is inferred.
    """

    idx = idx.copy()
    if freq is None:
        if idx.freq is None:
            freq = pd.infer_freq(idx)
        else:
            return idx
    idx.freq = pd.tseries.frequencies.to_offset(freq)
    if idx.freq is None:
        raise AttributeError(
            "no discernible frequency found to `idx`.  Specify"
            " a frequency string with `freq`."
        )
    return idx



import datetime
import email.utils
import re
from functools import partial

import Levenshtein
import networkx as nx
import numpy as np
import pandas as pd

from dbigbang.parse import get_date


def consolidate_senders_activity(activity_df, to_consolidate):
    """
    takes a DataFrame in the format returned by activity
    takes a list of tuples of format ('from 1', 'from 2') to consolidate
    returns the consolidated DataFrame (a copy, not in place)
    """

    df = activity_df.copy(deep=True)
    for consolidate in to_consolidate:
        column_a, column_b = consolidate
        if column_a in df.columns and column_b in df.columns:
            df[column_a] = df[column_a] + df[column_b]
            df.drop(column_b, inplace=True, axis=1)  # delete the second column
    return df


def matricize(series, func):
    """
    create a matrix by applying func to pairwise combos of elements in a Series
    returns a square matrix as a DataFrame
    should return a symmetric matrix if func(a,b) == func(b,a)
    should return the identity matrix if func == '=='
    """
    matrix = pd.DataFrame(columns=series, index=series)
    for index, element in enumerate(series):
        for second_index, second_element in enumerate(series):
            matrix.iloc[index, second_index] = func(element, second_element)

    return matrix


def minimum_but_not_self(column, dataframe):
    minimum = 100
    for index, value in dataframe[column].items():
        if index == column:
            continue
        if value < minimum:
            minimum = value
    return minimum


def sorted_matrix(from_dataframe, limit=None, sort_key=None):
    """
        Takes a dataframe with 'from' fields for column headers
    .
        Returns a sorted distance matrix for the column headers,
        using from_header_distance (see method).
    """
    if limit is None:
        limit = len(from_dataframe.columns)

    distancedf = matricize(
        from_dataframe.columns[: limit - 1], from_header_distance
    )

    # specify that the values in the matrix are integers
    df = distancedf.astype(int)

    # unless otherwise specified, sort to minimize the integer values with rows other than yourself
    sort_key = (
        sort_key
        if sort_key is not None
        else partial(minimum_but_not_self, dataframe=df)
    )

    new_columns = sorted(df.columns[: limit - 1], key=sort_key)
    new_df = df.reindex(index=new_columns, columns=new_columns)

    return new_df


def resolve_sender_entities(act, lexical_distance=0):
    """
    Given an Archive's activity matrix, return a dict of lists, each containing
    message senders ('From' fields) that have been groups to be
    probably the same entity.
    """

    # senders orders by descending total activity
    senders = act.sum(0).sort_values(ascending=False)

    return resolve_entities(
        senders, from_header_distance, threshold=lexical_distance
    )


def resolve_entities(significance, distance_function, threshold=0):
    """
    Takes a Series mapping entities (index) to significance (values, numerical).

    Resolves the entities based on a lexical distance function.

    Returns a dictionary of labeled (keys) entity lists (values).
    Key is the most significant member of the entity list.
    """
    entities = significance.index

    # entities in lexical order
    entities_lex = entities.sort_values()
    entities_lex_dict = dict([(p[1], p[0]) for p in enumerate(entities_lex)])

    n = len(entities)
    # binary matrix of similarity between entries
    sim = np.zeros((n, n))

    # find similarity
    for i in range(n):
        name = entities[i]
        i = entities_lex_dict[name]

        # checking only lexically close entries and
        # in proportion to total activity
        # is a performance hack.
        for j in range(i - (n - i + 1) // 2, i + (n - i + 1) // 2):
            d = distance_function(entities_lex[i], entities_lex[j])
            sim[i, j] = d <= threshold

    # An entity is a connected component of the resulting graph
    G = nx.Graph(sim)
    entities_list = [
        [entities_lex[j] for j in x] for x in nx.connected_components(G)
    ]

    # given each entity a label based on its most active 'member'
    entities_dict = {}
    for e in entities_list:
        # TODO: tighten up this labeling function
        label = sorted(e, key=lambda n: significance[n], reverse=True)[0]
        entities_dict[label] = e

    return entities_dict


ren = r"([\w\+\.\-]+(\@| at )[\w+\.\-]*) \((.*)\)"


def from_header_distance(a, b, verbose=False):
    """
    A distance measure specifically for the 'From' header of emails.
    Normalizes based on common differences in client handling of email,
    then computes Levenshtein distance between components of the field.
    """
    # this translate table is one way you are supposed to
    # delete characters from a unicode string
    stop_characters = str('"<>')
    stop_characters_map = dict((ord(char), None) for char in stop_characters)

    a_normal = ""
    b_normal = ""

    try:
        a_normal = (
            str(a).lower().translate(stop_characters_map).replace(" at ", "@")
        )
    except UnicodeDecodeError:
        a_normal = (
            a.decode("utf-8")
            .lower()
            .translate(stop_characters_map)
            .replace(" at ", "@")
        )

    try:
        b_normal = (
            str(b).lower().translate(stop_characters_map).replace(" at ", "@")
        )
    except UnicodeDecodeError:
        b_normal = (
            b.decode("utf-8")
            .lower()
            .translate(stop_characters_map)
            .replace(" at ", "@")
        )

    ag = re.match(ren, a_normal)
    bg = re.match(ren, b_normal)

    dist = float("inf")

    if ag is None or bg is None:
        if verbose:
            print("malformed pair:")
            print(a)
            print(b)
        dist = Levenshtein.distance(a_normal, b_normal)
    else:
        dist = Levenshtein.distance(
            ag.groups()[0], bg.groups()[0]
        ) + Levenshtein.distance(ag.groups()[1], bg.groups()[1])

        if len(ag.groups()[2]) > 5 and len(ag.groups()[2]) > 5:
            dist = min(
                dist, Levenshtein.distance(ag.groups()[2], bg.groups()[2])
            )

    return dist


def eij(m, parts, i, j):
    total_edges = m.sum().sum()

    part_i = parts[i]
    part_j = parts[j]

    edges_in_range = m[np.ix_(np.array(part_i), np.array(part_j))]

    return edges_in_range.sum().sum() / total_edges


def ai(m, parts, i):
    total = 0

    for j in range(len(parts)):
        total = total + eij(m, parts, i, j)

    return total


def bi(m, parts, i):
    total = 0

    for j in range(len(parts)):
        total = total + eij(m, parts, j, i)  # note switched i,j

    return total


def modularity(m, parts):
    """
    Compute modularity of an adjacency matrix.
    Use metric from:
        Zanetti, M. and Schweitzer, F. 2012.
        "A Network Perspective on Software Modularity"
        ARCS Workshops 2012, pp. 175-186.
    """

    expected = 0
    actual = 0

    for i in range(len(parts)):
        expected = expected + ai(m, parts, i) * bi(m, parts, i)
        actual = actual + eij(m, parts, i, i)

    q = (actual - expected) / (1 - expected)

    return q


def domain_name_from_email(name):
    address = email.utils.parseaddr(name)[1]
    if "@" in address:
        domain = address.split("@")[1]
    else:
        domain = address.split(" at ")[1]
    return domain.lower()


def containment_distance(a, b):
    """
    A case-insensitive distance measure on strings.

    Returns:
     - 0 if strings are identical
     - positive infinity if neither string contains
       the other
     - 1 / (minimum string length) if one string contains the other.

    Good for Organizations.
    I.e. "cisco" "Cisco" "Cisco Systems" are all 'close' (< .2)
    """
    a = a.lower()
    b = b.lower()

    base = 0

    if a == b:
        base = 0
    if a in b or b in a:
        base = 1
    else:
        base = float("inf")

    modifier = min(len(a), len(b))

    if modifier == 0:
        return float("inf")
    else:
        return base / modifier



import datetime
import logging
import mailbox

import numpy as np
import pandas as pd
import pytz

import dbigbang.process as process
from dbigbang.thread import Node, Thread
from config.config import CONFIG

from . import mailman, utils


def load(path):
    data = pd.read_csv(path)
    return Archive(data)


class ArchiveWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


class Archive(object):
    """A representation of a mailing list archive."""

    data = None
    activity = None
    threads = None
    entities = None

    def __init__(self, data, archive_dir=CONFIG.mail_path, mbox=False):
        """
        Initialize an Archive object.

        The behavior of the constructor depends on the type
        of its first argument, data.

        If data is a Pandas DataFrame, it is treated as a representation of
        email messages with columns for Message-ID, From, Date, In-Reply-To,
        References, and Body. The created Archive becomes a wrapper around a
        copy of the input DataFrame.

        If data is a string, then it is interpreted as a path to either a
        single .mbox file (if the optional argument single_file is True) or
        else to a directory of .mbox files (also in .mbox format). Note that
        the file extensions need not be .mbox; frequently they will be .txt.

        Upon initialization, the Archive object drops duplicate entries
        and sorts its member variable *data* by Date.
        """

        if isinstance(data, pd.core.frame.DataFrame):
            self.data = data.copy()
        elif isinstance(data, str):
            self.data = mailman.load_data(
                data, archive_dir=archive_dir, mbox=mbox
            )

        try:
            self.data["Date"] = pd.to_datetime(
                self.data["Date"],
                errors="coerce",
                infer_datetime_format=True,
                utc=True,
            )
        except Exception as e:
            logging.error(
                "Error while converting to datetime, despite coerce mode. "
                + f"The error message is: {e}"
            )
            raise ArchiveWarning(
                "Error while converting to datetime, despite coerce mode. "
                + f"The error message is: {e}"
            )

        try:
            self.data.drop_duplicates(inplace=True)
        except Exception as e:
            logging.error(
                "Error while removing duplicate messages, maybe timezone issues?"
                + f"The error message is: {e}",
                exc_info=True,
            )

        # Drops any entries with no Date field.
        # It may be wiser to optionally
        # do interpolation here.
        if self.data["Date"].isnull().any():
            # self.data.dropna(subset=['Date'], inplace=True)
            self.data = self.data[self.data["Date"].notnull()]
            # workaround for https://github.com/pandas-dev/pandas/issues/13407

        # convert any null fields to None -- csv saves these as nan sometimes
        self.data = self.data.where(pd.notnull(self.data), None)

        try:
            # set the index to be the Message-ID column
            self.data.set_index("Message-ID", inplace=True)
        except KeyError:
            # will get KeyError if Message-ID is already index
            pass

        # let's do a pass to try to find bad tzinfo's
        bad_indices = []
        for i, row in self.data.iterrows():
            try:
                row["Date"] < datetime.datetime.now(pytz.utc)
            except Exception as e:
                logging.error(
                    "Error timezone issues while detecting bad rows. "
                    + f"The error message is: {e}",
                    exc_info=True,
                )
                bad_indices.append(i)
                logging.info("Bad timezone on %s", row["Date"])
        if len(bad_indices) > 0:
            # drop those rows that threw an error
            self.data = self.data.drop(bad_indices)
            logging.info("Dropped %d rows", len(bad_indices))

        try:
            self.data.sort_values(by="Date", inplace=True)
        except Exception as e:
            logging.error(
                "Error while sorting, maybe timezone issues?"
                + f" The error message is: {e}",
                exc_info=True,
            )

        if self.data.empty:
            raise mailman.MissingDataException(
                "Archive after initial processing is empty. Was data collected properly?"
            )

    def resolve_entities(self, inplace=True):
        """Return data with resolved entities."""
        if self.entities is None:
            if self.activity is None:
                self.get_activity()

            self.entities = process.resolve_sender_entities(self.activity)

        to_replace = []

        value = []

        for e, names in list(self.entities.items()):
            for n in names:
                to_replace.append(n)
                value.append(e)

        data = self.data.replace(
            to_replace=to_replace, value=value, inplace=inplace
        )

        # clear and replace activity with resolved activity
        self.activity = None
        self.get_activity()

        if inplace:
            return self.data
        else:
            return data

    def get_activity(self, resolved=False):
        """
        Get the activity matrix of an Archive.

        Columns of the returned DataFrame are the Senders of emails.
        Rows are indexed by ordinal date.
        Cells are the number of emails sent by each sender on each data.

        If *resolved* is true, then default entity resolution is run on the
        activity matrix before it is returned.
        """
        if self.activity is None:
            self.activity = self.compute_activity(self)

        if resolved:
            self.entities = process.resolve_sender_entities(self.activity)
            eact = utils.repartition_dataframe(self.activity, self.entities)

            return eact

        return self.activity

    def compute_activity(self, clean=True):
        """Return the computed activity."""
        mdf = self.data

        if clean:
            # unnecessary?
            if mdf["Date"].isnull().any():
                mdf = mdf.dropna(subset=["Date"])

            mdf = mdf[
                mdf["Date"] < datetime.datetime.now(pytz.utc)
            ]  # drop messages apparently in the future

        mdf2 = mdf.reindex(columns=["From", "Date"])
        mdf2["Date"] = mdf["Date"].apply(lambda x: x.toordinal())

        activity = (
            mdf2.groupby(["From", "Date"]).size().unstack("From").fillna(0)
        )
        new_date_range = np.arange(mdf2["Date"].min(), mdf2["Date"].max())
        # activity.set_index('Date')
        activity = activity.reindex(new_date_range, fill_value=0)

        return activity

    def get_threads(self, verbose=False):
        """Get threads."""

        if self.threads is not None:
            return self.threads

        df = self.data

        threads = list()
        visited = dict()

        total = df.shape[0]
        c = 0

        for i in df.iterrows():

            if verbose:
                c += 1
                if c % 1000 == 0:
                    print("Processed %d of %d" % (c, total))

            if i[1]["In-Reply-To"] == "None":
                root = Node(i[0], i[1])
                visited[i[0]] = root
                threads.append(Thread(root))
            elif i[1]["In-Reply-To"] not in list(visited.keys()):
                root = Node(i[1]["In-Reply-To"])
                succ = Node(i[0], i[1], root)
                root.add_successor(succ)
                visited[i[1]["In-Reply-To"]] = root
                visited[i[0]] = succ
                threads.append(Thread(root, known_root=False))
            else:
                parent = visited[i[1]["In-Reply-To"]]
                node = Node(i[0], i[1], parent)
                parent.add_successor(node)
                visited[i[0]] = node

        self.threads = threads

        return threads

    def save(self, path, encoding="utf-8"):
        """Save data to csv file."""
        self.data.to_csv(path, ",", encoding=encoding)


def find_footer(messages, number=1):
    """
    Returns the footer of a DataFrame of emails.

    A footer is a string occurring at the tail of most messages.
    Messages can be a DataFrame or a Series
    """
    if isinstance(messages, pd.DataFrame):
        messages = messages["Body"]

    # sort in lexical order of reverse strings to maximize foot length
    srb = messages.apply(lambda x: None if x is None else x[::-1]).order()
    # srb = df.apply(lambda x: None if x['Body'] is None else x['Body'][::-1],
    #              axis=1).order()
    # begin walking down the series looking for maximal overlap
    counts = {}

    last = None

    def clean_footer(foot):
        return foot.strip()

    for b in srb:
        if last is None:
            last = b
            continue
        elif b is None:
            continue
        else:
            head, i = utils.get_common_head(b, last, delimiter="\n")
            head = clean_footer(head[::-1])
            last = b

            if head in counts:
                counts[head] = counts[head] + 1
            else:
                counts[head] = 1

        last = b

    # reduce candidates that are strictly longer and less frequent
    # than most promising footer candidates
    for n, foot1 in sorted(
        [(v, k) for k, v in list(counts.items())], reverse=True
    ):
        for foot2, m in list(counts.items()):
            if n > m and foot1 in foot2 and len(foot1) > 0:
                counts[foot1] = counts[foot1] + counts[foot2]
                del counts[foot2]

    candidates = sorted(
        [(v, k) for k, v in list(counts.items())], reverse=True
    )

    return candidates[0:number]



import sys
import warnings

from snakefood import gendeps

warnings.simplefilter("ignore", UserWarning)
sys.stdout = open("dependencies.txt", "w")


def main():
    gendeps.gendeps()


if __name__ == "__main__":
    main()



# twopeople.py
#
# Written by Raj Agrawal and Ki Deuk Kim
#
# Contains functions used to analyze communication between two people in mailing list
# Examples can be found in ipython notebook "Collaboration Robustness" in examples folder
# Each function needs a pandas DataFrame called "exchanges" that contains every two-pair
# communication between participants in a mailing list.

from pprint import pprint as pp

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import pytz

import dbigbang.graph as graph
import dbigbang.mailman as mailman
import dbigbang.parse as parse
import dbigbang.process as process
from dbigbang.archive import Archive


# Gets the target two people A, B to analyze and returns
# the amount of time they communicated in the mailing list
# in TimeDelta type
def duration(exchanges, A, B):
    AtoB = exchanges[exchanges["From_original"] == A]
    AtoB = AtoB[AtoB["From_response"] == B]
    BtoA = exchanges[exchanges["From_original"] == B]
    BtoA = BtoA[BtoA["From_response"] == A]
    if len(AtoB) == 0:
        return max(BtoA["Date"]) - min(BtoA["Date"])
    if len(BtoA) == 0:
        return max(AtoB["Date"]) - min(AtoB["Date"])
    return max(max(AtoB["Date"]), max(BtoA["Date"])) - min(
        min(AtoB["Date"]), min(BtoA["Date"])
    )


# Returns the number of replies that two people A and B sent to
# each other in a tuple (# of replies from A to B, # of replies from B to A)
def num_replies(exchanges, A, B):
    AtoB = exchanges[exchanges["From_original"] == A]
    AtoB = AtoB[AtoB["From_response"] == B]
    BtoA = exchanges[exchanges["From_original"] == B]
    BtoA = BtoA[BtoA["From_response"] == A]
    return (len(AtoB), len(BtoA))


# Returns the reciprocity of communication between two people A and B
# in float type. This expresses how interactively they communicated to each
# other
def reciprocity(exchanges, A, B):
    num = num_replies(exchanges, A, B)
    return float(min(num)) / max(num)


# Finds every unique pair (A, B) from the pandas DataFrame "exchanges"
# and returns them in set data type
def unique_pairs(exchanges):
    pairs = set()
    total_responses = len(exchanges["From_original"])
    for i in range(total_responses):
        pair = (exchanges["From_original"][i], exchanges["From_response"][i])
        pair_reversed = (
            exchanges["From_response"][i],
            exchanges["From_original"][i],
        )
        if pair_reversed not in pairs:
            pairs.add(pair)
    return pairs


# Forms a new Pandas DataFrame that contains information about communication
# between a pair A and B using functions provided above and returns the result
def panda_pair(exchanges, A, B):
    try:
        return pd.DataFrame(
            [
                {
                    "A": A,
                    "B": B,
                    "duration": duration(exchanges, A, B),
                    "num_replies": sum(num_replies(exchanges, A, B)),
                    "reciprocity": reciprocity(exchanges, A, B),
                }
            ]
        )
    except Exception:
        print('No exchange between "%s" and "%s" exists.' % (A, B))


# With given pairs of communication, returns a Pandas DataFrame that contains
# communication information between two people A and B in every pair
def panda_allpairs(exchanges, pairs):
    data_list = []
    for pair in pairs:
        A = pair[0]
        B = pair[1]
        data_list.append(
            {
                "A": A,
                "B": B,
                "duration": duration(exchanges, A, B),
                "num_replies": sum(num_replies(exchanges, A, B)),
                "reciprocity": reciprocity(exchanges, A, B),
            }
        )
    return pd.DataFrame(data_list)



import codecs
import datetime
import fnmatch
import gzip
import logging
import mailbox
import os
import re
import subprocess
import urllib.error
import urllib.parse
import urllib.request
import warnings
from pprint import pprint as pp

import pandas as pd
import yaml
from validator_collection import checkers

from dbigbang.parse import get_date
from config.config import CONFIG

from . import listserv, parse, w3crawl

ml_exp = re.compile(r"/([\w-]+)/?$")
txt_exp = re.compile(r'href="(\d\d\d\d-\w*\.txt)"')
gz_exp = re.compile(r'href="(\d\d\d\d-\w*\.txt\.gz)"')
ietf_ml_exp = re.compile(r'href="([\d-]+.mail)"')
w3c_archives_exp = re.compile(r"lists\.w3\.org")
listserv_archives_exp = re.compile(r"list\.etsi\.org")

mailing_list_path_expressions = [gz_exp, ietf_ml_exp, txt_exp]

PROVENANCE_FILENAME = "provenance.yaml"


class InvalidURLException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class MissingDataException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def load_data(
    name: str, archive_dir: str = CONFIG.mail_path, mbox: bool = False
):
    """
    Load the data associated with an archive name, given as a string.

    Attempt to open {archives-directory}/NAME.csv as data.

    Failing that, if the the name is a URL, it will try to derive
    the list name from that URL and load the .csv again.

    Args:
        name: archive name
        archive_dir: archive directory path
        mbox:
    """

    if mbox:
        return open_list_archives(name, archive_dir=archive_dir, mbox=True)

    # a first pass at detecting if the string is a URL...
    if not (name.startswith("http://") or name.startswith("https://")):
        path = os.path.join(archive_dir, name + ".csv")

        if os.path.exists(path):
            data = pd.read_csv(path)
            return data
        else:
            logging.warning("No data available at %s", path)
    else:
        path = os.path.join(archive_dir, get_list_name(name) + ".csv")

        if os.path.exists(path):
            data = pd.read_csv(path)
            return data
        else:
            logging.warning(
                "No data found for %s. Check directory name and whether archives have been collected.",
                name,
            )


def collect_from_url(
    url: str, archive_dir: str = CONFIG.mail_path, notes=None
):
    """Collect data from a given url."""

    url = url.rstrip()
    try:
        has_archives = collect_archive_from_url(
            url, archive_dir=archive_dir, notes=notes
        )
    except urllib.error.HTTPError:
        logging.exception("HTTP Error in collecting archive: %s", url)
        return None

    if has_archives:
        unzip_archive(url)
        try:
            data = open_list_archives(url)
        except MissingDataException as e:  # don't strictly need to open the archives during the collection process, so catch the Exception and return
            print(e)
            return None

        # hard coding the archives directory in too many places
        # need to push this default to a configuration file
        path = os.path.join(archive_dir, get_list_name(url) + ".csv").replace(
            "\\", "/"
        )

        try:
            data.to_csv(path, ",", encoding="utf-8")
        except Exception as e:
            print(e)
            # if encoding doesn't work...don't encode? ----better not not encode !!
            # try:
            #     data.to_csv(path, ",")
            # except Exception as e:
            #     print e
            #     print "Can't export data. Aborting."
            #     return None

        return data
    else:
        return None


def urls_to_collect(urls_file: str):
    """Collect urls given urls in a file."""
    urls = []
    for url in open(urls_file):
        url = url.strip()
        if url.startswith("#"):  # comment lines should be ignored
            continue
        if len(url) == 0:  # ignore empty lines
            continue
        if checkers.is_url(url) is not True:
            logging.warning("invalid url: %s" % url)
            continue
        urls.append(url)
    return urls


def collect_from_file(
    urls_file: str,
    archive_dir: str = CONFIG.mail_path,
    notes=None,
):
    """Collect urls from a file."""
    urls = urls_to_collect(urls_file)
    for url in urls:
        collect_from_url(url, archive_dir=archive_dir, notes=notes)


def get_list_name(url):
    """
    Return the 'list name' from a canonical mailman archive url.

    Otherwise return the same URL.
    """
    # TODO: it would be better to catch these non-url cases earlier
    url = url.rstrip()

    if ml_exp.search(url) is not None:
        return ml_exp.search(url).groups()[0]
    else:
        warnings.warn("No mailing list name found at %s" % url)
        return url


def normalize_archives_url(url):
    """
    Normalize url.

    will try to infer, find or guess the most useful archives URL, given a URL.

    Return normalized URL, or the original URL if no improvement is found.
    """
    # change new IETF mailarchive URLs to older, still available text .mail archives
    new_ietf_exp = re.compile(
        "https://mailarchive\\.ietf\\.org/arch/search/"
        "\\?email_list=(?P<list_name>[\\w-]+)"
    )
    ietf_text_archives = (
        r"https://www.ietf.org/mail-archive/text/\g<list_name>/"
    )
    new_ietf_browse_exp = re.compile(
        r"https://mailarchive.ietf.org/arch/browse/(?P<list_name>[\w-]+)/?"
    )

    match = new_ietf_exp.match(url)
    if match:
        return re.sub(new_ietf_exp, ietf_text_archives, url)

    match = new_ietf_browse_exp.match(url)
    if match:
        return re.sub(new_ietf_browse_exp, ietf_text_archives, url)

    return url  # if no other change found, return the original URL


def archive_directory(base_dir, list_name):
    """Archive a directory."""
    arc_dir = os.path.join(base_dir, list_name)
    if not os.path.exists(arc_dir):
        os.makedirs(arc_dir)
    return arc_dir


def populate_provenance(directory, list_name, list_url, notes=None):
    """Create a provenance metadata file for current mailing list collection."""
    provenance = {
        "list": {
            "list_name": list_name,
            "list_url": list_url,
        },
        "date_collected": str(datetime.date.today()),  # uses ISO-8601
        "complete": False,
        "code_version": {
            "long_hash": subprocess.check_output(
                ["git", "rev-parse", "HEAD"]
            ).strip(),
            "description": subprocess.check_output(
                ["git", "describe", "--long", "--always", "--all"]
            ).strip()
            # 'version': '' #TODO: programmatically access DBigBang version number, see #288
        },
    }
    if notes:
        provenance["notes"] = notes

    file_path = os.path.join(directory, PROVENANCE_FILENAME)
    if os.path.isfile(file_path):  # a provenance file already exists
        pass  # skip for now
    else:
        file_handle = open(file_path, "w")
        yaml.dump(provenance, file_handle)
        logging.info("Created provenance file for %s" % (list_name))
        file_handle.close()


def access_provenance(directory):
    """
    Return an object with provenance information located in the given directory,
    or None if no provenance was found.
    """
    file_path = os.path.join(directory, PROVENANCE_FILENAME)
    if os.path.isfile(file_path):  # a provenance file already exists
        file_handle = open(file_path, "r")
        provenance = yaml.load(file_handle)
        return provenance
    return None


def update_provenance(directory, provenance):
    """Update provenance file with given object."""
    file_path = os.path.join(directory, PROVENANCE_FILENAME)
    file_handle = open(file_path, "w")
    yaml.dump(provenance, file_handle)
    logging.info("Updated provenance file in %s", directory)
    file_handle.close()


def collect_archive_from_url(url, archive_dir=CONFIG.mail_path, notes=None):
    """
    Collect archives (generally tar.gz) files from mailmain archive page.

    Return True if archives were downloaded, False otherwise
    (for example if the page lists no accessible archive files).
    """
    list_name = get_list_name(url)
    logging.info("Getting archive page for %s", list_name)

    if w3c_archives_exp.search(url):
        return w3crawl.collect_from_url(url, archive_dir, notes=notes)
    elif listserv_archives_exp.search(url):
        listserv.ListservArchive.from_url(
            name="3GPP",
            url_root=url,
            url_home=url + "HOME",
            instant_dump=True,
        )

    response = urllib.request.urlopen(url)
    html = codecs.decode(response.read())

    results = []
    for exp in mailing_list_path_expressions:
        results.extend(exp.findall(html))

    pp(results)

    # directory for downloaded files
    arc_dir = archive_directory(archive_dir, list_name)

    populate_provenance(
        directory=arc_dir, list_name=list_name, list_url=url, notes=notes
    )

    encountered_error = False
    # download monthly archives
    for res in results:
        result_path = os.path.join(arc_dir, res)
        # this check is redundant with urlretrieve
        if not os.path.isfile(result_path):
            gz_url = "/".join([url.strip("/"), res])
            logging.info("retrieving %s", gz_url)
            resp = urllib.request.urlopen(gz_url)
            if resp.getcode() == 200:
                logging.info("200 - writing file to %s", result_path)
                output = open(result_path, "wb")
                output.write(resp.read())
                output.close()
            else:
                logging.warning(
                    "%s error code trying to retrieve %s"
                    % (str(resp.getcode(), gz_url))
                )
                encountered_error = True

    if (
        not encountered_error
    ):  # mark that all available archives were collected
        provenance = access_provenance(arc_dir)
        provenance["complete"] = True
        update_provenance(arc_dir, provenance)

    # return True if any archives collected, false otherwise
    return len(results) > 0


def unzip_archive(url, archive_dir=CONFIG.mail_path):
    """Unzip archive files."""
    arc_dir = archive_directory(archive_dir, get_list_name(url))

    gzs = [
        os.path.join(arc_dir, fn)
        for fn in os.listdir(arc_dir)
        if fn.endswith(".txt.gz")
    ]

    logging.info("Unzipping %d archive files", len(gzs))

    for gz in gzs:
        try:
            f = gzip.open(gz, "rb")
            content = codecs.decode(f.read())
            f.close()

            txt_fn = str(gz[:-3])

            f2 = open(txt_fn, "w")
            f2.write(content)
            f2.close()
        except Exception as e:
            print(e)


# This works for the names of the files. Order them.
# datetime.datetime.strptime('2000-November',"%Y-%B")

# This doesn't yet work for parsing the dates. Because of %z Bullshit
# datetime.datetime.strptime(arch[0][0].get('Date'),"%a, %d %b %Y %H:%M:%S %z")

# x is a String, a Message, or a list of Messages
# The payload of a Message may be a String, a Message, or a list of Messages.
# OR maybe it's never just a Message, but always a list of them.
def recursive_get_payload(x):
    """Get payloads recursively."""
    if isinstance(x, str):
        return x
    elif isinstance(x, list):
        # return [recursive_get_payload(y) for y in x]
        return recursive_get_payload(x[0])
    # TODO: define 'email'
    # elif isinstance(x, email.message.Message):
    #    return recursive_get_payload(x.get_payload())
    else:
        print(x)
        return None


def open_list_archives(
    url: str,
    archive_dir: str = CONFIG.mail_path,
    mbox: bool = False,
) -> pd.DataFrame:
    """
    Return a list of all email messages contained in the specified directory.

    TODO: this argument should be re-named. sometimes it's being called with
        actual URLs, other times with subdirectory names, leading to spurious
        warnings in various places.

    Args:
        url: the name of a subdirectory of the directory specified
            in argument *archive_dir*. This directory is expected to contain
            files with extensions .txt, .mail, or .mbox. These files are all
            expected to be in mbox format-- i.e. a series of blocks of text
            starting with headers (colon-separated key-value pairs) followed by
            an email body.
        archive_dir: directory containing all messages.

    Returns:
    """
    if (url is None) and (archive_dir is None):
        raise ValueError("Either `url` or `archive_dir` must be given.")

    messages = None

    if mbox and (os.path.isfile(os.path.join(archive_dir, url))):
        # treat string as the path to a file that is an mbox
        box = mailbox.mbox(os.path.join(archive_dir, url), create=False)
        messages = list(box.values())

    else:
        # assume string is the path to a directory with many
        list_name = get_list_name(url)
        arc_dir = archive_directory(archive_dir, list_name)

        file_extensions = [".txt", ".mail", ".mbox"]

        txts = [
            os.path.join(arc_dir, fn)
            for fn in os.listdir(arc_dir)
            if any([fn.endswith(extension) for extension in file_extensions])
        ]

        logging.info("Opening %d archive files", len(txts))

        def mbox_reader(stream):
            """Read a non-ascii message from mailbox"""
            data = stream.read()
            text = data.decode(encoding="utf-8", errors="replace")
            return mailbox.mboxMessage(text)

        arch = [
            mailbox.mbox(txt, factory=mbox_reader, create=False).values()
            for txt in txts
        ]

        if len(arch) == 0:
            raise MissingDataException(
                (
                    "No messages in %s under %s. Did you run the "
                    "collect_mail.py script?"
                )
                % (archive_dir, list_name)
            )

        messages = [item for sublist in arch for item in sublist]

    return messages_to_dataframe(messages)


def open_activity_summary(url, archive_dir=CONFIG.mail_path):
    """
    Open the message activity summary for a particular mailing list (as specified by url).

    Return the dataframe, or return None if no activity summary export file is found.
    """
    list_name = get_list_name(url)
    arc_dir = archive_directory(archive_dir, list_name)

    activity_csvs = fnmatch.filter(os.listdir(arc_dir), "*-activity.csv")
    if len(activity_csvs) == 0:
        logging.info("No activity summary found for %s." % list_name)
        return None

    if len(activity_csvs) > 1:
        logging.info(
            "Multiple possible activity summary files found for %s, using the first one."
            % list_name
        )

    activity_csv = activity_csvs[0]
    path = os.path.join(arc_dir, activity_csv)
    activity_frame = pd.read_csv(path, index_col=0, encoding="utf-8")
    return activity_frame


def get_text(msg):
    """Get text from a message."""
    ## This code for character detection and dealing with exceptions is terrible
    ## It is in need of refactoring badly. - sb
    import chardet

    text = ""
    if msg.is_multipart():
        html = None
        for part in msg.walk():
            charset = part.get_content_charset()
            if part.get_content_type() == "text/plain":
                try:
                    text = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    )
                except LookupError:
                    logging.debug(
                        "Unknown encoding %s in message %s. Will use UTF-8 instead.",
                        charset,
                        msg["Message-ID"],
                    )
                    charset = "utf-8"
                    text = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    )
            if part.get_content_type() == "text/html":
                try:
                    html = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    )
                except LookupError:
                    logging.debug(
                        "Unknown encoding %s in message %s. Will use UTF-8 instead.",
                        charset,
                        msg["Message-ID"],
                    )
                    charset = "utf-8"
                    html = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    )

        if text is not None:
            return text.strip()
        else:
            import html2text

            h = html2text.HTML2Text()
            h.encoding = "utf-8"
            return str(h.handle(html))
    else:
        charset = msg.get_content_charset() or "utf-8"
        if charset != "utf-8":
            logging.debug("charset is %s" % (charset))
        text = msg.get_payload()
        return text.strip()


def messages_to_dataframe(messages):
    """
    Turn a list of parsed messages into a dataframe of message data,
    indexed by message-id, with column-names from headers.
    """
    # extract data into a list of tuples -- records -- with
    # the Message-ID separated out as an index
    # valid_messages = [m for m in messages if m.get()

    pm = [
        (
            m.get("Message-ID"),
            str(m.get("From")).replace("\\", " "),
            str(m.get("Subject")),
            get_date(m),
            str(m.get("In-Reply-To")),
            str(m.get("References")),
            get_text(m),
        )
        for m in messages
        if m.get("From")
    ]

    mdf = pd.DataFrame.from_records(
        list(pm),
        index="Message-ID",
        columns=[
            "Message-ID",
            "From",
            "Subject",
            "Date",
            "In-Reply-To",
            "References",
            "Body",
        ],
    )
    mdf.index.name = "Message-ID"

    return mdf



import ast
import fnmatch
import json
import os
import re
import subprocess
import sys

import networkx as nx
import pandas as pd
import requests
from IPython.nbconvert import PythonExporter
from IPython.nbformat import current as nbformat

from config.config import CONFIG

from .git_repo import GitRepo, MultiGitRepo

repoLocation = CONFIG.repo_path
examplesLocation = CONFIG.urls_path
nameRegex = re.compile(r"([^/]*)\.git$")
fileRegex = re.compile(r".*\/(.*)")


class RepoLoaderWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


def repo_already_exists(filepath):
    return os.path.exists(filepath)


def url_to_name(url):
    """
    Converts a github url (e.g. https://github.com/sbenthall/dbigbang.git) to
    a human-readable name (dbigbang) by looking at the word between the last "/" and ".git".
    """
    url = url.replace("\n","")
    name = nameRegex.search(url).group(1)
    return name


def name_to_filepath(name):
    """
    Converts a name of a repo to its filepath.
    Currently, these go to ../archives/sample_git_repos/{name}/
    """
    newLoc = repoLocation + name
    return newLoc


def filepath_to_name(filepath):
    """
    Converts a filepath (../archives/sample_git_repos/{name}) to a name.
    Note that this will fail if the filepath ends in a "/". It must end
    in the name of the folder.
    Thus, it should be ../archives/sample_git_repos/{name} not
    ../archives/sample_git_repos/{name}/
    """
    name = fileRegex.search(filepath).group(1)
    return name


def create_graph(dic):
    """
    Converts a dictionary of dependencies into a  NetworkX DiGraph.
    """
    G = nx.DiGraph()
    for f in dic:
        for dependency in dic[f]:
            G.add_edge(f, dependency)
    return G


def get_files(filepath):
    """
    Returns a list of the Python files in a directory, and
    converts IPython notebooks into Python source code and
    includes them with the Python files.
    """
    os.chdir(filepath)
    files = []
    for root, dirnames, filenames in os.walk("."):
        for filename in fnmatch.filter(filenames, "*.py"):
            files.append(os.path.join(root, filename))

        for filename in fnmatch.filter(filenames, "*.ipynb"):
            try:
                with open(filename) as fh:
                    nb = nbformat.reads_json(fh.read())
                    export_path = filename.replace(".ipynb", ".py")
                    exporter = PythonExporter()
                    source, meta = exporter.from_notebook_node(nb)

                with open(export_path, "w+") as fh:
                    fh.writelines(source)
                files.append()
            except Exception as e:
                raise RepoLoaderWarning(
                    f"May have issues with JSON encoding. The error message is: {e}"
                )
    return files


def get_dependency_network(filepath):
    """
    Given a directory, collects all Python and IPython files and
    uses the Python AST to create a dictionary of dependencies from them.
    Returns the dependencies converted into a NetworkX graph.
    """
    files = get_files(filepath)
    dependencies = {}
    for file in set(files):
        tree = ast.parse(file)
        for node in tree.getChildren()[1].nodes:
            if isinstance(node, tree.Import):
                if file in dependencies:
                    dependencies[file].append(node.names[0][0])
                else:
                    dependencies[file] = [node.names[0][0]]

            elif isinstance(node, tree.From):
                if file in dependencies:
                    dependencies[file].append(
                        node.modname + "/" + node.names[0][0]
                    )
    return create_graph(dependencies)


def get_repo(repo_in, in_type="name", update=False):
    """
    Takes three different options for type:
        - remote: basically a git url
        - name (default): a name like 'scipy' which the method can expand to a url
        - local: a filepath to a file on the local system
            (basically an existing git directory on this computer)
    This returns an initialized GitRepo object with its data and name already loaded.
    """
    # Input is name
    if in_type == "name":
        filepath = name_to_filepath(repo_in)
        ans = None
        if not update:
            ans = get_cache(repo_in)
        if ans is not None:
            return ans
        print(("Checking for " + str(repo_in) + " at " + str(filepath)))
        ans = get_repo(filepath, "local", update)

        if isinstance(ans, GitRepo):
            # We cache it hopefully???
            ans.commit_data.to_csv(
                cache_path(repo_in), sep="\t", encoding="utf-8"
            )
        else:
            print("We failed to find a local copy of this repo")
        return ans

    # Input is a local file
    if in_type == "local":
        if repo_already_exists(repo_in):
            name = filepath_to_name(repo_in)
            return GitRepo(url=repo_in, name=name)
        else:
            print(("Invalid filepath: " + repo_in))
            return None

    if in_type == "remote":
        name = url_to_name(repo_in)
        filepath = name_to_filepath(name)
        if not repo_already_exists(filepath):
            print("Cloning the repo from remote")
            fetch_repo(repo_in)
        return get_repo(name, "name", update)

    else:
        print("Invalid input")  # TODO: Clarify this error
        return None


def fetch_repo(url):
    """
    Takes in a git url and uses shell commands
    to clone the git repo into sample_git_repos/

    TODO: We shouldn't use this with shell=True because of security concerns.
    """
    # TODO: We are repeatedly calculating name and filepath
    url = url.replace("\n", "")
    name = url_to_name(url)
    newLoc = name_to_filepath(name)
    command = ["git " + "clone " + url + " " + newLoc]
    subprocess.call(command, shell=True)


def cache_path(name):
    """
    Takes in a name (dbigbang)
    Returns where its cached file should be (../sample_git_repos/dbigbang_backup.csv)
    """
    return repoLocation + str(name) + "_backup.csv"


def get_cache(name):
    """
    Takes in a name (dbigbang)
    Returns a GitRepo object containing the cache data if the cache exists
    Returns None otherwise.
    """
    filepath = cache_path(name)
    if os.path.exists(filepath):
        c = pd.read_csv(filepath, sep="\t", encoding="utf-8")
        fp = name_to_filepath(name)
        ans = GitRepo(name=name, url=fp, cache=c)
        return ans
    return None


def get_multi_repo(repo_names=None, repos=None):
    """
    As of now, this only accepts names/repos, not local urls
    TODO: This could be optimized
    """
    if repos is None:
        repos = list()
        for name in repo_names:
            repo = get_repo(name, in_type="name")
            repos.append(repo)

    for repo in repos:
        repo.commit_data["Repo Name"] = repo.name

    ans = MultiGitRepo(repos)
    return ans


def load_org_repos(org_name):
    """
    fetches a list of all repos in an organization from github
    and gathers their URL's (of the form *.git)
    It dumps these into ../examples/{org_name}_urls.txt
    """
    github_url = "https://api.github.com/orgs/" + org_name + "/repos"
    r = requests.get(github_url)
    data = r.json()
    urls = []
    for repo in data:
        if "git_url" in repo:
            urls.append(repo["git_url"])
    if len(urls) == 0:
        print(("Found no repos in group: " + str(org_name)))
        return None
    else:
        addr = examplesLocation + str(org_name) + "_urls.txt"
        f = open(addr, "w")
        f.write("\n".join(urls))
        print(("Wrote git urls to " + addr))
        return urls


def get_org_repos(org_name):
    """
    Checks to see if we have the urls for a given org
    If we don't, it fetches them.
    Once we do, it returns a list of GitRepo objects from the urls.
    """
    addr = examplesLocation + str(org_name) + "_urls.txt"
    urls = None
    if not os.path.isfile(addr):
        urls = load_org_repos(org_name)
    else:
        urls = open(addr, "r")
    ans = list()
    for url in urls:
        ans.append(get_repo(url, "remote"))

    return ans


def get_org_multirepo(org_name):
    repos = get_org_repos(org_name)
    ans = get_multi_repo(repos=repos)
    return ans



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

config = {
    'description': 'Analysis of Mailman archives',
    'long_description': 'Analysis of Mailman archives',
    'long_description_content_type': 'text/x-rst',
    'author': 'Sebastian Benthall',
    'url': 'https://gitbub.com/datactive/dbigbang',
    'download_url': 'https://gitbub.com/datactive/dbigbang',
    'author_email': 'sb@ischool.berkeley.edu',
    'version': '0.2',
    'install_requires': [
        'beautifulsoup4',
        'chardet',
        'coverage',
        'html2text',
        'ipython',
        'jinja2',
        'jsonschema',
        'matplotlib',
        'networkx',
        'numpy',
        'pandas',
        'pytest',
        'python-dateutil',
        'python-Levenshtein',
        'pytz',
        'pyzmq',
        'tornado',
        'requests',
        'pyyaml',
        'common',
        'powerlaw'],
    'packages': ['dbigbang'],
    'scripts': [],
    'name': 'DBigBang'}

setup(**config)



import email
import logging
import re
import warnings
from pprint import pprint as pp

import dateutil.parser as dp
import pytz

re_cache = {
    "top_exp": re.compile(r"From .*\d\d\d\d\n"),
    "msg_id": re.compile(r"<\S*@\S*>"),
}


def split_references(refs):
    return re_cache["msg_id"].findall(refs)


def get_refs(refs):
    return re_cache["msg_id"].findall(refs)


def clean_mid(mid):
    try:
        return get_refs(mid)[0]
    except IndexError:
        print(mid)
        return mid


def clean_from(m_from):
    """
    Return a person's name extracted from 'From' field
    of email, based on heuristics.
    """

    cleaned = m_from

    try:
        if "(" in m_from:
            cleaned = m_from[m_from.index("(") + 1 : m_from.rindex(")")]
        elif "<" in m_from:
            # if m_from.index("<") > -1:
            cleaned = m_from[0 : m_from.index("<") - 1]

    except ValueError:
        warnings.warn("%s is hard to clean" % (m_from))

    cleaned = cleaned.strip('"')

    return cleaned


def normalize_email_address(address):
    """
    Takes a valid email address and returns a normalized one, for matching purposes.
    """
    # TODO: drop the + labeling
    return address.lower()


def clean_name(name):
    """
    Clean just the name portion from email.utils.parseaddr.

    Returns None if the name portion is missing anything name-like. Otherwise, returns the cleaned name.
    """

    # we see these specific strings due to parsing issues in email somewhere
    name = name.replace("unknown charset", " ")
    name = name.replace("wrong string", " ")
    name = name.replace("_", " ")

    # these are stop characters we can just delete
    stop_characters = str("\"<>\\()/:?%!+'@")
    stop_characters_map = dict((ord(char), None) for char in stop_characters)

    name = name.translate(stop_characters_map)

    # do we need to also catch email archives that use anti-spam measures?
    # like: .replace(' at ','@')

    # TODO: decode or collapse rfc2231 encodings, like '=?utf-8?q?carlos_gonz=c3=a1lez-cadenas?=' ?

    name = name.strip()  # remove leading and trailing whitespace

    if len(name) > 0:
        return name
    else:
        return None


def tokenize_name(clean_name):
    """
    Create a tokenized version of a name, good for comparison and sorting for entity resolution.

    Takes a Unicode name already cleaned of most punctuation and spurious characters, hopefully.
    """

    # make lower case, remove "." and ",", tokenize and lexicographically sort the tokens, join by spaces, return as a string

    stop_characters = str('".,')
    stop_characters_map = dict((ord(char), None) for char in stop_characters)
    name = clean_name.translate(stop_characters_map)

    tokens = name.lower().split()  # splits on whitespace

    if len(tokens) == 0:
        return None

    tokenized_name = " ".join(sorted(tokens))

    return tokenized_name


def guess_first_name(cleaned_from):
    """
    Attempt to extract a person's first name from the cleaned version of their name
    (from a 'From' field). This may or may not be the given name. Returns None
    if heuristic doesn't recognize a separable first name.
    """

    cleaned_from = (
        cleaned_from.strip()
    )  # remove leading and trailing whitespace

    # accomodate Last, First name format
    if "," in cleaned_from:
        parts = cleaned_from.split(",")
        if len(parts) > 2:
            return None
        first_part = parts[1].strip()

        if (
            " " in first_part
        ):  # Last, First Middle? Or something entirely different
            return None
        else:
            return first_part

    elif " " in cleaned_from:
        parts = cleaned_from.split(" ")
        if len(parts) == 2:  # e.g. First Last
            return parts[0]
        if len(parts) == 3:  # e.g. First Middle Last
            return parts[0]
        return None
    else:  # no spaces or commas? with a single name, more likely to be a handle than a given name
        return None


def get_date(message):
    try:
        # some mail clients add a parenthetical timezone
        ds = str(message.get("Date"))
        ds = re.sub(r"\(.*$", "", ds)
        ds = re.sub(r"--", "-", ds)
        ds = re.sub(r" Hora.*$", "", ds)

        date = dp.parse(ds)

        # this adds noise and could raise trouble
        if date.tzinfo is None:
            date = pytz.utc.localize(date)

        return date
    except (TypeError, ValueError):
        logging.debug("Date parsing error on: %s", ds)
        return None
